from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from . import component_base

__all__ = ["Grid"]


@dataclass
class GridChildPosition:
    row: int
    column: int
    width: int = 1
    height: int = 1


class Grid(component_base.FundamentalComponent):
    """
    A container which arranges its children in a table-like grid.

    Grids arrange their children in a table-like grid. Each child is placed in
    one or more cells of the grid.

    Example:
    ```python
    # TODO
    ```

    Attributes:
        row_spacing: The amount of space between rows of the grid.

        column_spacing: The amount of space between columns of the grid.
    """

    _: KW_ONLY
    row_spacing: float
    column_spacing: float

    # These must be annotated, otherwise rio won't understand that grids have
    # child components and won't copy over the new values when two Grids are
    # reconciled
    _children: List[rio.Component]
    _child_positions: List[GridChildPosition]

    def __init__(
        self,
        *rows: Iterable[rio.Component],
        row_spacing: float = 0.0,
        column_spacing: float = 0.0,
        key: Optional[str] = None,
        margin: Optional[float] = None,
        margin_x: Optional[float] = None,
        margin_y: Optional[float] = None,
        margin_left: Optional[float] = None,
        margin_top: Optional[float] = None,
        margin_right: Optional[float] = None,
        margin_bottom: Optional[float] = None,
        width: Union[Literal["natural", "grow"], float] = "natural",
        height: Union[Literal["natural", "grow"], float] = "natural",
        align_x: Optional[float] = None,
        align_y: Optional[float] = None,
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.row_spacing = row_spacing
        self.column_spacing = column_spacing

        # JS can only work with lists of Components, so we'll store the
        # components and their positions separately
        _children: List[rio.Component] = []
        _child_positions: List[GridChildPosition] = []

        for row_nr, row_components in enumerate(rows):
            for col_nr, component in enumerate(row_components):
                _children.append(component)
                _child_positions.append(
                    GridChildPosition(
                        row=row_nr,
                        column=col_nr,
                    )
                )

        self._children = _children
        self._child_positions = _child_positions

        self._explicitly_set_properties_.update(["_children", "_child_positions"])

    def add_child(
        self,
        child: rio.Component,
        row: int,
        column: int,
        width: int = 1,
        height: int = 1,
    ) -> None:
        self._children.append(child)
        self._child_positions.append(GridChildPosition(row, column, width, height))

    def _custom_serialize(self) -> JsonDoc:
        return {
            "_children": [child._id for child in self._children],
            "_child_positions": [vars(pos) for pos in self._child_positions],
        }


Grid._unique_id = "Grid-builtin"



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from . import component_base

__all__ = [
    "Card",
]


class Card(component_base.FundamentalComponent):
    """
    A container that visually encompasses its child components.

    Cards are used to group related components together, and to visually
    separate them from other components, allowing you to display them in a
    structured way.

    Cards are also often used as large buttons. They can be configured to
    elevate slightly when the mouse hovers over them, indicating to the user
    that they support interaction.

    Attributes:
        child: The component to display inside the card.

        corner_radius: The radius of the card's corners. If set to `None`, it
            is picked from the active theme.

        on_press: An event handler that is called when the card is clicked.
            Note that attaching an even handler will also modify the appearance
            of the card, to signal the possible interaction to the user. See
            `elevate_on_hover` and `colorize_on_hover` for details.

        elevate_on_hover: Whether the card should elevate slightly when the
            mouse hovers over it. If set to `None` the card will elevate if
            an `on_press` event handler is attached.

        colorize_on_hover: Whether the card should change its color when the
            mouse hovers over it. If set to `None` the card will change its
            color if an `on_press` event handler is attached.
    """

    child: rio.Component
    _: KW_ONLY
    corner_radius: Union[None, float, Tuple[float, float, float, float]] = None
    on_press: rio.EventHandler[[]] = None
    elevate_on_hover: Optional[bool] = None
    colorize_on_hover: Optional[bool] = None
    inner_margin: bool = True
    color: rio.ColorSet = "neutral"

    async def _on_message(self, msg: Any) -> None:
        # Trigger the press event
        await self.call_event_handler(self.on_press)

        # Refresh the session
        await self.session._refresh()

    def _custom_serialize(self) -> JsonDoc:
        thm = self.session.theme
        color = thm._serialize_colorset(self.color)

        report_press = self.on_press is not None

        return {
            "corner_radius": self.corner_radius,
            "reportPress": report_press,
            "elevate_on_hover": (
                report_press if self.elevate_on_hover is None else self.elevate_on_hover
            ),
            "colorize_on_hover": (
                report_press
                if self.colorize_on_hover is None
                else self.colorize_on_hover
            ),
            "color": color,
        }


Card._unique_id = "Card-builtin"



from __future__ import annotations

import asyncio
import os
import socket
import sys
import threading
import webbrowser
from datetime import timedelta
from pathlib import Path
from typing import *  # type: ignore

import __main__
import fastapi
import uvicorn

import rio

from . import app_server, assets, debug
from .common import ImageLike

# Only available with the `window` extra
try:
    import webview  # type: ignore
except ImportError:
    webview = None


__all__ = [
    "App",
]


def _validate_build_function(
    build_function: Callable[[], rio.Component]
) -> Callable[[], rio.Component]:
    assert callable(
        build_function
    ), f"The App requires a function that returns a component, not {build_function!r}"

    def wrapper():
        component = build_function()

        assert isinstance(
            component, rio.Component
        ), f"The `build` function passed to the App must return a `Component` instance, not {component!r}."

        return component

    return wrapper


def default_connection_lost_message() -> rio.Component:
    return rio.Html(
        """
<div class="error-box">
    <svg xmlns="http://www.w3.org/2000/svg"
        style="fill: var(--rio-global-danger-bg); width: 1.6rem; height: 1.6rem;" viewBox="0 -960 960 960">
        <path
            d="M479.928-274.022q16.463 0 27.398-10.743 10.935-10.743 10.935-27.206 0-16.464-10.863-27.518-10.862-11.055-27.326-11.055-16.463 0-27.398 11.037-10.935 11.037-10.935 27.501 0 16.463 10.863 27.224 10.862 10.76 27.326 10.76Zm3.247-158.739q14.499 0 24.195-9.821 9.695-9.82 9.695-24.244v-188.935q0-14.424-9.871-24.244-9.871-9.821-24.369-9.821-14.499 0-24.195 9.821-9.695 9.82-9.695 24.244v188.935q0 14.424 9.871 24.244 9.871 9.821 24.369 9.821Zm-2.876 358.74q-84.202 0-158.041-31.879t-129.159-87.199q-55.32-55.32-87.199-129.201-31.878-73.88-31.878-158.167t31.878-158.2q31.879-73.914 87.161-128.747 55.283-54.832 129.181-86.818 73.899-31.986 158.205-31.986 84.307 0 158.249 31.968 73.942 31.967 128.756 86.768 54.815 54.801 86.79 128.883 31.976 74.083 31.976 158.333 0 84.235-31.986 158.07t-86.818 128.942q-54.833 55.107-128.873 87.169-74.04 32.063-158.242 32.063Zm.201-68.131q140.543 0 238.946-98.752 98.402-98.752 98.402-239.596 0-140.543-98.215-238.946-98.215-98.402-239.753-98.402-140.163 0-238.945 98.215-98.783 98.215-98.783 239.753 0 140.163 98.752 238.945 98.752 98.783 239.596 98.783ZM480-480Z" />
    </svg>
    <div style="font-weight: bold;">Connection lost</div>
</div>
<style>
    #rio-connection-lost-popup .error-box {
        position: absolute;
        top: 2rem;
        left: 50%;
        transform: translateX(-50%);
        width: unset;
        height: unset;

        display: flex;
        align-items: center;
        gap: 0.6rem;

        color: var(--rio-global-danger-bg);
        background: var(--rio-global-neutral-bg);
        padding: 1.5rem 2rem;
        border-radius: 99999px;
        box-shadow: 0 0 1rem var(--rio-global-shadow-color);
    }
</style>
        """
    )


class App:
    """
    Contains all the information needed to run a Rio app.

    Apps group all the information needed for Rio to run your application, such
    as its name, icon and, and the pages it contains. Apps also expose several
    lifetime events that you can use to perform tasks such as initialization and
    cleanup.

    If you're serving your app as a website, all users share the same `App`
    instance. If running in a window, there's only one window, and thus `App`,
    anyway.

    A basic setup may look like this:

    ```py
    rio_app = rio.App(
        name="My App",
        build=MyAppRoot,
    )
    ```

    You can then run this app, either as a local application in a window:

    ```py
    rio_app.run_in_window()
    ```

    Or you can create and run a webserver:

    ```py
    rio_app.run_as_web_server()
    ```

    Or create a server, without running it. This allows you to start the script
    externally with tools such as uvicorn:

    ```py
    fastapi_app = rio_app.as_fastapi()
    ```

    Attributes:
        name: The name to display for this app. This can show up in window
            titles, error messages and wherever else the app needs to be
            referenced in a nice, human-readable way.

        pages: The pages that make up this app. You can navigate between these
            using `Session.navigate_to` or using `Link` components. If running
            as website the user can also access these pages directly via their
            URL.

        assets_dir: The directory where the app's assets are stored. This allows
            you to conveniently access any images or other files that are needed
            by your app.
    """

    # Type hints so the documentation generator knows which fields exist
    name: str
    pages: Tuple[rio.Page, ...]
    assets_dir: Path

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        build: Optional[Callable[[], rio.Component]] = None,
        icon: Optional[ImageLike] = None,
        pages: Iterable[rio.Page] = tuple(),
        on_app_start: rio.EventHandler[App] = None,
        on_session_start: rio.EventHandler[rio.Session] = None,
        on_session_end: rio.EventHandler[rio.Session] = None,
        default_attachments: Iterable[Any] = (),
        ping_pong_interval: Union[int, float, timedelta] = timedelta(seconds=50),
        assets_dir: Union[str, os.PathLike, None] = None,
        theme: Union[rio.Theme, Tuple[rio.Theme, rio.Theme], None] = None,
        build_connection_lost_message: Callable[
            [], rio.Component
        ] = default_connection_lost_message,
    ):
        """
        Args:
            build: A function that returns the root component of the app. This
                function will be called whenever a new session is created. Note
                that since classes are callable in Python, you can pass a class
                here instead of a function, so long as the class doesn't require
                any arguments.

                If no build method is passed, the app will create a `PageView`
                as the root component.

            name: The name to display for this app. This can show up in window
                titles, error messages and wherever else the app needs to be
                referenced in a nice, human-readable way. If not specified,
                `Rio` name will try to guess a name based on the name of the
                main Python file.

            icon: The icon to display for this app. This can show up in window
                the title bars of windows, browser tabs, or similar.

            pages: The pages that make up this app. You can navigate between
                these using `Session.navigate_to` or using `Link` components. If
                running as website the user can also access these pages directly
                via their URL.

            on_app_start: A function that will be called when the app is first
                started. You can use this to perform any initialization tasks
                that need to happen before the app is ready to use.

                The app start will be delayed until this function returns. This
                makes sure initialization is complete before the app is
                displayed to the user. If you would prefer to perform
                initialization in the background try using `asyncio.create_task`
                to run your code in a separate task.

            on_session_start: A function that will be called each time a new
                session is created. In the context of a website that would be
                each time a new user visits the site. In the context of a window
                there is only one session, so this will only be called once.

                This function does not block the creation of the session. This
                is to make sure initialization code doesn't accidentally make
                the user wait.

            on_session_end: A function that will be called each time a session
                ends. In the context of a website that would be each time a user
                closes their browser tab. In the context of a window this will
                only be called once, when the window is closed.

            default_attachments: A list of attachments that will be attached to
                every new session.

            ping_pong_interval: Rio periodically sends ping-pong messages
                between the client and server to prevent overzealous proxies
                from closing the connection. The default value should be fine
                for most deployments, but feel free to change it if your hosting
                provider deploys a particularly obnoxious proxy.

            assets_dir: The directory where the app's assets are stored. This
                allows you to conveniently access any images or other files that
                are needed by your app. If not specified, Rio will assume the
                assets are stored in a directory called "assets" in the same
                directory as the main Python file.
        """
        main_file = _get_main_file()

        if name is None:
            name = _get_default_app_name(main_file)

        if build is None:
            build = rio.PageView

        if theme is None:
            theme = rio.Theme.from_color()

        self.name = name
        self._build = _validate_build_function(build)
        self._icon = None if icon is None else assets.Asset.from_image(icon)
        self.pages = tuple(pages)
        self._on_app_start = on_app_start
        self._on_session_start = on_session_start
        self._on_session_end = on_session_end
        self._default_attachments = tuple(default_attachments)
        self.assets_dir = main_file.parent / (
            "assets" if assets_dir is None else assets_dir
        )
        self._theme = theme
        self._build_connection_lost_message = build_connection_lost_message

        if isinstance(ping_pong_interval, timedelta):
            self._ping_pong_interval = ping_pong_interval
        else:
            self._ping_pong_interval = timedelta(seconds=ping_pong_interval)

    def _as_fastapi(
        self,
        *,
        running_in_window: bool,
        validator_factory: Optional[Callable[[rio.Session], debug.Validator]],
        internal_on_app_start: Optional[Callable[[], None]],
    ) -> fastapi.FastAPI:
        """
        Internal equivalent of `as_fastapi` that takes additional arguments.
        """
        return app_server.AppServer(
            self,
            running_in_window=running_in_window,
            on_session_start=self._on_session_start,
            on_session_end=self._on_session_end,
            default_attachments=self._default_attachments,
            validator_factory=validator_factory,
            internal_on_app_start=internal_on_app_start,
        )

    def as_fastapi(self):
        """
        Return a FastAPI instance that serves this app.

        This method returns a FastAPI instance that serves this app. This allows
        you to run the app with a custom server, such as uvicorn:

        ```py
        rio_app = rio.App(
            name="My App",
            build=MyAppRoot,
        )

        fastapi_app = rio_app.as_fastapi()
        ```

        You can then run this app with uvicorn:

        ```sh
        uvicorn my_app:fastapi_app
        ```
        """
        return self._as_fastapi(
            running_in_window=False,
            validator_factory=None,
            internal_on_app_start=None,
        )

    def _run_as_web_server(
        self,
        *,
        host: str,
        port: int,
        quiet: bool,
        validator_factory: Optional[Callable[[rio.Session], debug.Validator]],
        internal_on_app_start: Optional[Callable[[], None]],
    ) -> None:
        """
        Internal equivalent of `run_as_web_server` that takes additional
        arguments.
        """
        port = _ensure_valid_port(host, port)

        # Suppress stdout messages if requested
        kwargs = {}

        if quiet:
            kwargs["log_config"] = {
                "version": 1,
                "disable_existing_loggers": True,
                "formatters": {},
                "handlers": {},
                "loggers": {},
            }

        # Create the FastAPI server
        fastapi_app = self._as_fastapi(
            running_in_window=False,
            validator_factory=validator_factory,
            internal_on_app_start=internal_on_app_start,
        )

        # Serve
        uvicorn.run(
            fastapi_app,
            host=host,
            port=port,
            **kwargs,
        )

    def run_as_web_server(
        self,
        *,
        host: str = "localhost",
        port: int = 8000,
        quiet: bool = False,
    ) -> None:
        """
        Creates and runs a webserver that serves this app.

        This method creates and immediately runs a webserver that serves this
        app. This is the simplest way to run a Rio app.

        ```py
        rio_app = rio.App(
            name="My App",
            build=MyAppRoot,
        )

        rio_app.run_as_web_server()
        ```

        The will synchronously block until the server is shut down.

        Args:
            host: Which IP address to serve the webserver on. `localhost` will
                make the service only available on your local machine. This is
                the recommended setting if running behind a proxy like nginx.

            port: Which port the webserver should listen to.

            quiet: If `True` Rio won't send any routine messages to `stdout`.
                Error messages will be printed regardless of this setting.
        """
        self._run_as_web_server(
            host=host,
            port=port,
            quiet=quiet,
            validator_factory=None,
            internal_on_app_start=None,
        )

    def run_in_browser(
        self,
        *,
        host: str = "localhost",
        port: Optional[int] = None,
        quiet: bool = False,
    ) -> None:
        """
        Runs an internal webserver and opens the app in the default browser.

        This method creates and immediately runs a webserver that serves this
        app, and then opens the app in the default browser. This is a quick and easy
        way to access your app.

        ```py
        rio_app = rio.App(
            name="My App",
            build=MyAppRoot,
        )

        rio_app.run_in_browser()
        ```

        Args:
            host: Which IP address to serve the webserver on. `localhost` will
                make the service only available on your local machine. This is
                the recommended setting if running behind a proxy like nginx.

            port: Which port the webserver should listen to. If not specified,
                Rio will choose a random free port.

            quiet: If `True` Rio won't send any routine messages to `stdout`.
                Error messages will be printed regardless of this setting.
        """
        port = _ensure_valid_port(host, port)

        def on_startup() -> None:
            webbrowser.open(f"http://{host}:{port}")

        self._run_as_web_server(
            host=host,
            port=port,
            quiet=quiet,
            validator_factory=None,
            internal_on_app_start=on_startup,
        )

    def run_in_window(
        self,
        quiet: bool = True,
    ) -> None:
        """
        Runs the app in a local window.

        This method creates a window and displays the app in it. This is great
        if you don't want the complexity of running a web server, or wish to
        package your app as a standalone executable.

        ```py
        rio_app = rio.App(
            name="My App",
            build=MyAppRoot,
        )

        rio_app.run_in_window()
        ```

        This method requires the `window` extra. If you don't have it installed,
        you can install it with:

        ```sh
        pip install rio-ui[window]
        ```

        This method will synchronously block until the window is closed.  <!-- TODO is that correct? -->

        Args:
            quiet: If `True` Rio won't send any routine messages to `stdout`.
                Error messages will be printed regardless of this setting.
        """

        if webview is None:
            raise Exception(
                "The `window` extra is required to use `App.run_in_window`. Run `pip install rio[window]` to install it."
            )

        # Unfortunately, WebView must run in the main thread, which makes this
        # tricky. We'll have to banish uvicorn to a background thread, and shut
        # it down when the window is closed.

        host = "localhost"
        port = _ensure_valid_port(host, None)
        url = f"http://{host}:{port}"

        # This lock is released once the server is running
        lock = threading.Lock()
        lock.acquire()

        server: Optional[uvicorn.Server] = None

        # Start the server, and release the lock once it's running
        def run_web_server():
            fastapi_app = self._as_fastapi(
                running_in_window=True,
                validator_factory=None,
                internal_on_app_start=lock.release,
            )

            # Suppress stdout messages if requested
            log_level = "error" if quiet else "info"

            nonlocal server
            config = uvicorn.Config(
                fastapi_app,
                host=host,
                port=port,
                log_level=log_level,
                timeout_graceful_shutdown=1,  # Without a timeout, sometimes the server just deadlocks
            )
            server = uvicorn.Server(config)

            asyncio.run(server.serve())

        server_thread = threading.Thread(target=run_web_server)
        server_thread.start()

        # Wait for the server to start
        lock.acquire()

        # Start the webview
        try:
            webview.create_window(
                self.name,
                url,
            )
            webview.start()

        finally:
            assert isinstance(server, uvicorn.Server)

            server.should_exit = True
            server_thread.join()


def _get_main_file() -> Path:
    try:
        return Path(__main__.__file__)
    except AttributeError:
        pass

    # Find out if we're being executed by uvicorn
    main_file = Path(sys.argv[0])
    if (
        main_file.name != "__main__.py"
        or main_file.parent != Path(uvicorn.__file__).parent
    ):
        return main_file

    # Find out from which module uvicorn imported the app
    try:
        app_location = next(arg for arg in sys.argv[1:] if ":" in arg)
    except StopIteration:
        return main_file

    module_name, _, _ = app_location.partition(":")
    module = sys.modules[module_name]

    if module.__file__ is None:
        return main_file

    return Path(module.__file__)


def _get_default_app_name(main_file: Path) -> str:
    name = main_file.stem
    if name in ("main", "__main__", "__init__"):
        name = main_file.absolute().parent.stem

    return name.replace("_", " ").title()


def _choose_free_port(host: str) -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.bind((host, 0))
        return sock.getsockname()[1]


def _ensure_valid_port(host: str, port: Optional[int]) -> int:
    if port is None:
        return _choose_free_port(host)

    # if _port_is_in_use(host, port):
    #     raise ValueError(f"The port {host}:{port} is already in use")

    return port



from .app_root import *
from .banner import *
from .button import *
from .card import *
from .color_picker import *
from .column import *
from .component_base import *
from .container import *
from .custom_button import *
from .drawer import *
from .dropdown import *
from .flow_container import *
from .grid import *
from .html import *
from .icon import *
from .image import *
from .key_event_listener import *
from .labeled_column import *
from .link import *
from .list_items import *
from .list_view import *
from .markdown_view import *
from .media_player import *
from .mouse_event_listener import *
from .number_input import *
from .overlay import *
from .page_view import *
from .plot import *
from .popup import *
from .progress_bar import *
from .progress_circle import *
from .rectangle import *
from .revealer import *
from .row import *
from .scroll_container import *
from .scroll_target import *
from .slider import *
from .slideshow import *
from .spacer import *
from .stack import *
from .style_context import *
from .switch import *
from .switcher_bar import *
from .table import *
from .text import *
from .text_input import *
from .website import *



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import *  # type: ignore

import rio

from . import component_base

__all__ = [
    "Banner",
]


class Banner(component_base.Component):
    """
    Displays a short message to the user.

    Banners can either show a short text message to the users, or disappear
    entirely if no message is set. Use them to inform the user about the result
    of an action, to give feedback on their input, or anything else that needs
    to be communicated.

    The messages have one of four levels: success, info, warning, and error. The
    levels control the appearance of the notification bar, and allow you to
    quickly communicate the nature of the message to the user.

    Attributes:
        text: The text to display. If empty, the banner will disappear entirely.

        style: Controls the appearance of the banner. The style is one of
            `info`, `success`, `warning` and `danger`. Depending on the value
            the banner may change its colors and icon.

        multiline: Whether long text may be wrapped over multiple lines.
            Multiline banners are also styled slightly differently to make the
            icon fit their larger size.
    """

    text: str
    style: Literal["info", "success", "warning", "danger"]

    _: KW_ONLY
    markup: bool = False
    multiline: bool = False

    def build(self) -> rio.Component:
        # Early out: Nothing to show
        if not self.text:
            return rio.Column()

        # Prepare the style
        if self.style == "info":
            style_name = "secondary"
            icon = "info"
        elif self.style == "success":
            style_name = "success"
            icon = "check-circle"
        elif self.style == "warning":
            style_name = "warning"
            icon = "warning"
        elif self.style == "danger":
            style_name = "danger"
            icon = "error"
        else:
            raise ValueError(f"Invalid style: {self.style}")

        # Prepare the text child
        if self.markup:
            text_child = rio.MarkdownView(
                self.text,
                width="grow",
            )
        else:
            text_child = rio.Text(
                self.text,
                width="grow",
                multiline=self.multiline,
            )

        # Build the result
        if self.multiline:
            return rio.Card(
                child=rio.Row(
                    rio.Icon(
                        icon,
                        width=2.5,
                        height=2.5,
                        align_y=0,
                    ),
                    text_child,
                    spacing=1.5,
                    margin=0.5,
                ),
                color=style_name,
            )

        return rio.Card(
            child=rio.Row(
                rio.Icon(icon),
                text_child,
                spacing=0.8,
                align_x=0.5,
            ),
            color=style_name,
            corner_radius=self.session.theme.corner_radius_small,
        )



from __future__ import annotations

import asyncio
import collections
import enum
import inspect
import json
import logging
import secrets
import shutil
import time
import traceback
import typing
import weakref
from dataclasses import dataclass
from datetime import tzinfo
from pathlib import Path
from typing import *  # type: ignore

import aiofiles
import babel
import introspection
import unicall
import uniserde
from uniserde import Jsonable, JsonDoc

import rio

from . import (
    app_server,
    assets,
    color,
    common,
    errors,
    font,
    global_state,
    inspection,
    routing,
    self_serializing,
    text_style,
    theme,
    user_settings_module,
)
from .components import component_base

__all__ = ["Session"]


T = typing.TypeVar("T")


@dataclass
class ComponentData:
    build_result: rio.Component

    # Keep track of how often this component has been built. This is used by
    # components to determine whether they are still part of their parent's current
    # build output.
    build_generation: int


class WontSerialize(Exception):
    pass


async def dummy_send_message(message: Jsonable) -> None:
    raise NotImplementedError()


async def dummy_receive_message() -> JsonDoc:
    raise NotImplementedError()


def _host_and_get_fill_as_css_variables(
    fill: rio.FillLike,
    sess: "Session",
) -> Dict[str, str]:
    # Convert the fill
    fill = rio.Fill._try_from(fill)

    if isinstance(fill, rio.SolidFill):
        return {
            "color": f"#{fill.color.hex}",
            "background": "none",
            "background-clip": "unset",
            "fill-color": "unset",
        }

    assert isinstance(fill, (rio.LinearGradientFill, rio.ImageFill)), fill
    return {
        "color": "var(--rio-local-text-color)",
        "background": fill._as_css_background(sess),
        "background-clip": "text",
        "fill-color": "transparent",
    }


class SessionAttachments:
    def __init__(self, sess: Session):
        self._session = sess
        self._attachments: Dict[type, object] = {}

    def __iter__(self) -> Iterator[object]:
        return iter(self._attachments.values())

    def __contains__(self, typ: type) -> bool:
        return typ in self._attachments

    def __getitem__(self, typ: Type[T]) -> T:
        """
        Retrieves an attachment from this session.

        Attached values are:
        - the session's user settings
        - any value that was previously attached using `Session.attach`
        """
        try:
            return self._attachments[typ]  # type: ignore
        except KeyError:
            raise KeyError(typ) from None

    def _add(self, value: Any, synchronize: bool) -> None:
        # User settings need special care
        if isinstance(value, user_settings_module.UserSettings):
            # Get the previously attached value, and unlink it from the session
            try:
                old_value = self[type(value)]
            except KeyError:
                pass
            else:
                old_value._rio_session_ = None

            # Link the new value to the session
            value._rio_session_ = self._session

            # Trigger a resync
            if synchronize:
                self._session.create_task(
                    value._synchronize_now(self._session),
                    name="write back user settings (entire settings instance changed)",
                )

        # Save it with the rest of the attachments
        self._attachments[type(value)] = value

    def add(self, value: Any) -> None:
        """
        Attaches the given value to the `Session`. It can be retrieved later
        using `session.attachments[...]`.
        """
        self._add(value, synchronize=True)


class Session(unicall.Unicall):
    """
    Represents a single client connection to the app.

    A session corresponds to a single connection to a client. It maintains all
    state related to this client including local settings, the currently active
    page, and others.

    Sessions are created automatically by the app and should not be created
    manually.
    """

    def __init__(
        self,
        app_server_: app_server.AppServer,
        session_token: str,
        base_url: rio.URL,
        initial_page: rio.URL,
    ):
        super().__init__(
            send_message=dummy_send_message,
            receive_message=dummy_receive_message,
        )

        self._app_server = app_server_
        self._session_token = session_token

        # Components need unique ids, but we don't want them to be globally unique
        # because then people could guesstimate the approximate number of
        # visitors on a server based on how quickly the component ids go up. So
        # each session will assign its components ids starting from 0.
        self._next_free_component_id: int = 0

        # Keep track of the last time we successfully communicated with the
        # client. After a while with no communication we close the session.
        self._last_interaction_timestamp = time.monotonic()

        # The URL to the app's root/index page. The current page url will always
        # start with this.
        assert base_url.is_absolute(), base_url
        self._base_url = base_url

        # The current page. This is publicly accessible as property
        assert initial_page.is_absolute(), initial_page
        self._active_page_url: rio.URL = initial_page

        # Also the current page url, but as stack of the actual page instances
        self._active_page_instances: Tuple[rio.Page, ...] = ()

        # Keeps track of running asyncio tasks. This is used to make sure that
        # the tasks are cancelled when the session is closed.
        self._running_tasks: Set[asyncio.Task[object]] = set()

        # Maps all PageViews to the id of the PageView one higher up in the
        # component tree. Root PageViews are mapped to None. This map is kept up
        # to date by PageViews themselves.
        self._page_views: weakref.WeakKeyDictionary[
            rio.PageView, Optional[int]
        ] = weakref.WeakKeyDictionary()

        # All components / methods which should be called when the session's
        # page has changed.
        #
        # The methods don't have the component bound yet, so they don't unduly
        # prevent the component from being garbage collected.
        self._page_change_callbacks: weakref.WeakKeyDictionary[
            rio.Component, Tuple[Callable[[rio.Component], None], ...]
        ] = weakref.WeakKeyDictionary()

        # All components / methods which should be called when the session's
        # window size has changed.
        self._on_window_resize_callbacks: weakref.WeakKeyDictionary[
            rio.Component, Callable[[rio.Component], None]
        ] = weakref.WeakKeyDictionary()

        # All fonts which have been registered with the session. This maps the
        # name of the font to the font's assets, which ensures that the assets
        # are kept alive until the session is closed.
        self._registered_font_assets: Dict[str, List[assets.Asset]] = {
            font.ROBOTO.name: [],
            font.ROBOTO_MONO.name: [],
        }

        # These are injected by the app server after the session has already been created
        self._root_component: rio.Component
        self.external_url: Optional[str]  # None if running in a window
        self.preferred_locales: Tuple[
            babel.Locale, ...
        ]  # Always has at least one member
        self.timezone: tzinfo

        self.window_width: float
        self.window_height: float

        self.theme: theme.Theme

        # Must be acquired while synchronizing the user's settings
        self._settings_sync_lock = asyncio.Lock()

        # If `running_in_window`, this contains all the settings loaded from the
        # json file. We need to keep this around so that we can update the
        # settings that have changed and write everything back to the file.
        self._settings_json: Dict[str, object] = {}

        # Weak dictionaries to hold additional information about components.
        # These are split in two to avoid the dictionaries keeping the
        # components alive. Notice how both dictionaries are weak on the actual
        # component.
        #
        # Never access these directly. Instead, use helper functions
        # - `lookup_component`
        # - `lookup_component_data`
        self._weak_components_by_id: weakref.WeakValueDictionary[
            int, rio.Component
        ] = weakref.WeakValueDictionary()

        self._weak_component_data_by_component: weakref.WeakKeyDictionary[
            rio.Component, ComponentData
        ] = weakref.WeakKeyDictionary()

        # Keep track of all dirty components, once again, weakly.
        #
        # Components are dirty if any of their properties have changed since the
        # last time they were built. Newly created components are also considered
        # dirty.
        #
        # Use `register_dirty_component` to add a component to this set.
        self._dirty_components: weakref.WeakSet[rio.Component] = weakref.WeakSet()

        # HTML components have source code which must be evaluated by the client
        # exactly once. Keep track of which components have already sent their
        # source code.
        self._initialized_html_components: Set[str] = set(
            inspection.get_child_component_containing_attribute_names_for_builtin_components()
        )

        # This lock is used to order state updates that are sent to the client.
        # Without it a message which was generated later might be sent to the
        # client before an earlier message, leading to invalid component
        # references.
        self._refresh_lock = asyncio.Lock()

        # All components wich are currently in the widget tree and are
        # subscribed to either `on_mount` or `on_unmount` events.
        self._watched_and_mounted_components: Set[rio.Component] = set()

        # Attachments. These are arbitrary values which are passed around inside
        # of the app. They can be looked up by their type.
        # Note: These are initialized by the AppServer.
        self.attachments = SessionAttachments(self)

        # This allows easy access to the app's assets. Users can simply write
        # `component.session.assets / "my_asset.png"`.
        self.assets = self._app_server.app.assets_dir

    @property
    def app(self) -> rio.App:
        """
        Returns the app which this session belongs to.
        """
        return self._app_server.app

    @property
    def running_in_window(self) -> bool:
        """
        Returns `True` if the app is running in a local window, and `False` if
        it is hosted as a website.
        """
        return self._app_server.running_in_window

    @property
    def running_as_website(self) -> bool:
        """
        Returns `True` if the app is running as a website, and `False` if it is
        running in a local window.
        """
        return self._app_server.running_in_window

    @property
    def base_url(self) -> rio.URL:
        """
        Returns the base URL of the app.

        Only available when running as a website.
        """
        if self._app_server.running_in_window:
            raise RuntimeError(
                "Cannot get the base URL of an app that is running in a window"
            )

        return self._base_url

    @property
    def active_page_url(self) -> rio.URL:
        """
        Returns the current page as a tuple of strings.

        This property is read-only. To change the page, use `Session.navigate_to`.
        """
        return self._active_page_url

    @property
    def active_page_instances(self) -> Tuple[rio.Page, ...]:
        """
        Returns the current page as a tuple of `Page` instances.

        This property is read-only. To change the page, use `Session.navigate_to`.
        """
        return self._active_page_instances

    def close(self) -> None:
        self.create_task(self._close())

    async def _close(self) -> None:
        try:
            del self._app_server._active_session_tokens[self._session_token]
        except KeyError:
            # This can happen if two _close() tasks are running at the same
            # time. Just abort in that case.
            return

        # Fire the session end event
        await self._call_event_handler(self._app_server.on_session_end, self)

        if self.running_in_window:
            window = await self._get_webview_window()
            window.destroy()

            # Save all of the settings
            for attachment in self.attachments:
                if isinstance(attachment, user_settings_module.UserSettings):
                    await attachment._synchronize_now(self)
        else:
            try:
                await self._remote_close_session()
            except RuntimeError:  # Websocket is already closed
                pass

        # Cancel all running tasks
        for task in self._running_tasks:
            task.cancel()

    async def _get_webview_window(self):
        import webview  # type: ignore

        # The window may not have opened yet, so we'll wait until it exists
        while True:
            window = webview.active_window()

            if window is not None:
                return window

            await asyncio.sleep(0.2)

    @overload
    async def _call_event_handler(
        self,
        handler: common.EventHandler[[]],
    ) -> None:
        ...

    @overload
    async def _call_event_handler(
        self,
        handler: common.EventHandler[[T]],
        event_data: T,
        /,
    ) -> None:
        ...

    async def _call_event_handler(
        self,
        handler: common.EventHandler[...],
        *event_data: object,
    ) -> None:
        # Event handlers are optional
        if handler is None:
            return

        # If the handler is available, call it and await it if necessary
        try:
            result = handler(*event_data)

            if inspect.isawaitable(result):
                await result

        # Display and discard exceptions
        except Exception:
            print("Exception in event handler:")
            traceback.print_exc()

    def create_task(
        self,
        coro: Coroutine[Any, None, T],
        *,
        name: Optional[str] = None,
    ) -> asyncio.Task[T]:
        """
        Creates an `asyncio.Task` that is cancelled when the session is closed.

        This is identical to `asyncio.create_task`, except that any tasks are
        automatically cancelled when the session is closed. This makes sure that
        old tasks don't keep piling up long after they are no longer needed.

        Args:
            coro: The coroutine to run.

            name: An optional name for the task. Assigning descriptive names can
                be helpful when debugging.
        """
        task = asyncio.create_task(coro, name=name)

        self._running_tasks.add(task)
        task.add_done_callback(self._running_tasks.remove)

        return task

    def navigate_to(
        self,
        target_url: Union[rio.URL, str],
        *,
        replace: bool = False,
    ) -> None:
        """
        Switches the app to display the given page URL.

        Switches the app to display the given page URL. If `replace` is `True`,
        the browser's most recent history entry is replaced with the new page.
        This means that the user can't go back to the previous page using the
        browser's back button. If `False`, a new history entry is created,
        allowing the user to go back to the previous page.

        Args:
            target_url: The URL of the page to navigate to.

            replace: If `True`, the browser's most recent history entry is
                replaced with the new page. If `False`, a new history entry is
                created, allowing the user to go back to the previous page.
        """

        # Normalize the target URL. Having it always be lowercase helps the user
        # avoid navigation problems because of casing issues.
        if isinstance(target_url, rio.URL):
            target_url = str(target_url).lower()

        target_url = rio.URL(target_url)

        # Determine the full page to navigate to
        target_url_absolute = self.active_page_url.join(target_url)

        # Is this a page, or a full URL to another site?
        try:
            common.make_url_relative(
                self._base_url,
                target_url_absolute,
            )

        # This is an external URL. Navigate to it
        except ValueError:

            async def history_worker() -> None:
                await self._evaluate_javascript(
                    f"""
window.location.href = {json.dumps(str(target_url))};
""",
                )

            self.create_task(history_worker(), name="history worker")
            return

        # Is any guard opposed to this page?
        active_page_instances, active_page = routing.check_page_guards(
            self, target_url_absolute
        )

        # Update the current page
        self._active_page_url = active_page
        self._active_page_instances = tuple(active_page_instances)

        # Dirty all PageViews to force a rebuild
        for page_view in self._page_views:
            self._register_dirty_component(
                page_view, include_children_recursively=False
            )

        self.create_task(self._refresh())

        # Update the browser's history
        async def history_worker() -> None:
            method = "replaceState" if replace else "pushState"
            await self._evaluate_javascript(
                f"""
window.history.{method}(null, null, {json.dumps(str(target_url))})
window.scrollTo({{ top: 0, behavior: 'smooth' }});
""",
            )

        self.create_task(history_worker())

        # Trigger the `on_page_change` event
        async def event_worker() -> None:
            for component, callbacks in self._page_change_callbacks.items():
                for callback in callbacks:
                    self.create_task(
                        self._call_event_handler(callback, component),
                        name="`on_page_change` event handler",
                    )

        self.create_task(event_worker())

    def _register_dirty_component(
        self,
        component: rio.Component,
        *,
        include_children_recursively: bool,
    ) -> None:
        """
        Add the component to the set of dirty components. The component is only held
        weakly by the session.

        If `include_children_recursively` is true, all children of the component
        are also added.

        The children of non-fundamental components are not added, since they will
        be added after the parent is built anyway.
        """
        self._dirty_components.add(component)

        if not include_children_recursively or not isinstance(
            component, component_base.FundamentalComponent
        ):
            return

        for child in component._iter_direct_children():
            self._register_dirty_component(
                child,
                include_children_recursively=True,
            )

    def _refresh_sync(self) -> Tuple[Dict[rio.Component, bool], Set[rio.Component]]:
        """
        See `refresh` for details on what this function does.

        The refresh process must be performed atomically, without ever yielding
        control flow to the async event loop. TODO WHY

        To make sure async isn't used unintentionally this part of the refresh
        function is split into a separate, synchronous function.

        The session keeps track of components which are no longer referenced in its
        component tree. Those components are NOT included in the function's result.
        """

        # Keep track of all components which are visited. Only they will be sent to
        # the client.
        visited_components: collections.Counter[rio.Component] = collections.Counter()

        # Build all dirty components
        while self._dirty_components:
            component = self._dirty_components.pop()

            # Remember that this component has been visited
            visited_components[component] += 1

            # Catch deep recursions and abort
            build_count = visited_components[component]
            if build_count > 5:
                raise RecursionError(
                    f"The component `{component}` has been rebuilt {build_count} times during a single refresh. This is likely because one of your components' `build` methods is modifying the component's state"
                )

            # Fundamental components require no further treatment
            if isinstance(component, component_base.FundamentalComponent):
                continue

            # Others need to be built
            global_state.currently_building_component = component
            global_state.currently_building_session = self

            try:
                build_result = component.build()
            finally:
                global_state.currently_building_component = None
                global_state.currently_building_session = None

            # Sanity check
            if not isinstance(build_result, rio.Component):  # type: ignore[unnecessary-isinstance]
                raise ValueError(
                    f"The output of `build` methods must be instances of `rio.Component`, but `{component}` returned `{build_result}`"
                )

            # Has this component been built before?
            try:
                component_data = self._weak_component_data_by_component[component]

            # No, this is the first time
            except KeyError:
                # Create the component data and cache it
                component_data = ComponentData(build_result, 0)
                self._weak_component_data_by_component[component] = component_data

            # Yes, rescue state. This will:
            #
            # - Look for components in the build output which correspond to
            #   components in the previous build output, and transfers state
            #   from the new to the old component ("reconciliation")
            #
            # - Replace any references to new, reconciled components in the
            #   build output with the old components instead
            #
            # - Add any dirty components from the build output (new, or old but
            #   changed) to the dirty set.
            #
            # - Update the component data with the build output resulting from
            #   the operations above
            else:
                self._reconcile_tree(component, component_data, build_result)

                # Increment the build generation
                component_data.build_generation = global_state.build_generation
                global_state.build_generation += 1

                # Reconciliation can change the build result. Make sure nobody
                # uses `build_result` instead of `component_data.build_result`
                # from now on.
                del build_result

            # Inject the builder and build generation
            weak_builder = weakref.ref(component)

            children = component_data.build_result._iter_direct_and_indirect_child_containing_attributes(
                include_self=True,
                recurse_into_high_level_components=False,
            )
            for child in children:
                child._weak_builder_ = weak_builder
                child._build_generation_ = component_data.build_generation

        # Determine which components are alive, to avoid sending references to
        # dead components to the frontend.
        alive_cache: Dict[rio.Component, bool] = {
            self._root_component: True,
        }

        return alive_cache, {
            component
            for component in visited_components
            if component._is_in_component_tree(alive_cache)
        }

    async def _refresh(self) -> None:
        """
        Make sure the session state is up to date. Specifically:

        - Call build on all components marked as dirty
        - Recursively do this for all freshly spawned components
        - mark all components as clean

        Afterwards, the client is also informed of any changes, meaning that
        after this method returns there are no more dirty components in the
        session, and Python's state and the client's state are in sync.
        """

        # For why this lock is here see its creation in `__init__`
        async with self._refresh_lock:
            # Refresh and get a set of all components which have been visited
            alive_cache, visited_components = self._refresh_sync()

            # Avoid sending empty messages
            if not visited_components:
                return

            # Serialize all components which have been visited
            delta_states: Dict[int, JsonDoc] = {
                component._id: self._serialize_and_host_component(component)
                for component in visited_components
            }

            await self._update_component_states(visited_components, delta_states)

            # Trigger the `on_unmount` event
            #
            # Notes:
            # - All events are triggered only after the client has been notified
            #   of the changes. This way, if the event handlers trigger another
            #   client message themselves, any referenced widgets will already
            #   exist
            for comp in list(self._watched_and_mounted_components):
                # Still alive
                if comp._is_in_component_tree(alive_cache):
                    continue

                # Trigger the event
                for handler in comp._rio_event_handlers_[rio.event.EventTag.ON_UNMOUNT]:
                    self.create_task(
                        self._call_event_handler(handler, comp),
                        name="`on_unmount` event handler",
                    )

                # Remember that this component is no longer mounted
                self._watched_and_mounted_components.remove(comp)

            # Trigger the `on_mount` event
            #
            # Notes:
            # - See the note on running after notifying the client above
            # - This function enlarges the set of watched components. The
            #   `on_unmount` event handler iterates over that set. By processing
            #   that handler first it only needs to process a smaller set
            for component in visited_components - self._watched_and_mounted_components:
                # Disinterested
                if not component._watch_tree_mount_and_unmount_:
                    continue

                # Trigger the event
                for handler in component._rio_event_handlers_[
                    rio.event.EventTag.ON_MOUNT
                ]:
                    self.create_task(
                        self._call_event_handler(handler, component),
                        name="`on_mount` event handler",
                    )

                # Remember that this component is now mounted
                self._watched_and_mounted_components.add(component)

    async def _update_component_states(
        self, visited_components: Set[rio.Component], delta_states: Dict[int, JsonDoc]
    ) -> None:
        # Initialize all new FundamentalComponents
        for component in visited_components:
            if (
                not isinstance(component, component_base.FundamentalComponent)
                or type(component)._unique_id in self._initialized_html_components
            ):
                continue

            await component._initialize_on_client(self)
            self._initialized_html_components.add(type(component)._unique_id)

        # Check whether the root component needs replacing
        if self._root_component in visited_components:
            root_component_id = self._root_component._id
        else:
            root_component_id = None

        # Send the new state to the client
        await self._remote_update_component_states(delta_states, root_component_id)

    async def _send_reconnect_message(self) -> None:
        self._initialized_html_components.clear()

        # For why this lock is here see its creation in `__init__`
        async with self._refresh_lock:
            visited_components: Set[component_base.Component] = set()
            delta_states = {}

            for component in self._root_component._iter_component_tree():
                visited_components.add(component)
                delta_states[component._id] = self._serialize_and_host_component(
                    component
                )

            await self._update_component_states(visited_components, delta_states)

    def _serialize_and_host_value(
        self,
        value: Any,
        type_: type,
    ) -> Jsonable:
        """
        Which values are serialized for state depends on the annotated
        datatypes. There is no point in sending fancy values over to the client
        which it can't interpret.

        This function attempts to serialize the value, or raises a
        `WontSerialize` exception if this value shouldn't be included in the
        state.
        """
        origin = typing.get_origin(type_)
        args = typing.get_args(type_)

        # Explicit check for some types. These don't play nice with `isinstance` and
        # similar methods
        if origin is Callable:
            raise WontSerialize()

        if isinstance(type_, typing.TypeVar):
            raise WontSerialize()

        # Basic JSON values
        if type_ in (bool, int, float, str, None):
            return value

        # Enums
        if inspect.isclass(type_) and issubclass(type_, enum.Enum):
            return uniserde.as_json(value, as_type=type_)

        # Sequences of serializable values
        if origin is list:
            return [self._serialize_and_host_value(v, args[0]) for v in value]

        # ColorSet
        # Important: This must happen before the SelfSerializing check, because
        # `value` might be a `Color`
        if origin is Union and set(args) == color._color_set_args:
            thm = self.theme
            return thm._serialize_colorset(value)

        # Self-Serializing
        if isinstance(value, self_serializing.SelfSerializing):
            return value._serialize(self)

        # Optional
        if origin is Union and len(args) == 2 and type(None) in args:
            if value is None:
                return None

            type_ = next(type_ for type_ in args if type_ is not type(None))
            return self._serialize_and_host_value(value, type_)

        # Literal
        if origin is Literal:
            return self._serialize_and_host_value(value, type(value))

        # Components
        if introspection.safe_is_subclass(type_, rio.Component):
            return value._id

        # Invalid type
        raise WontSerialize()

    def _serialize_and_host_component(
        self,
        component: rio.Component,
    ) -> JsonDoc:
        """
        Serializes the component, non-recursively. Children are serialized just by
        their `_id`.

        Non-fundamental components must have been built, and their output cached in
        the session.
        """
        result: JsonDoc = {
            "_python_type_": type(component).__name__,
        }

        # Add layout properties, in a more succinct way than sending them
        # separately
        result["_margin_"] = (
            component.margin_left,
            component.margin_top,
            component.margin_right,
            component.margin_bottom,
        )
        result["_size_"] = (
            component.width if isinstance(component.width, (int, float)) else None,
            component.height if isinstance(component.height, (int, float)) else None,
        )
        result["_align_"] = (
            component.align_x,
            component.align_y,
        )
        result["_grow_"] = (
            component.width == "grow",
            component.height == "grow",
        )

        # If it's a fundamental component, serialize its state because JS needs it.
        # For non-fundamental components, there's no reason to send the state to
        # the frontend.
        if isinstance(component, component_base.FundamentalComponent):
            for name, type_ in inspection.get_attributes_to_serialize(
                type(component)
            ).items():
                try:
                    result[name] = self._serialize_and_host_value(
                        getattr(component, name),
                        type_,
                    )
                except WontSerialize:
                    pass

        # Encode any internal additional state. Doing it this late allows the custom
        # serialization to overwrite automatically generated values.
        if isinstance(component, component_base.FundamentalComponent):
            result["_type_"] = component._unique_id
            result.update(component._custom_serialize())

        else:
            # Take care to add underscores to any properties here, as the
            # user-defined state is also added and could clash
            result["_type_"] = "Placeholder"
            result["_child_"] = self._weak_component_data_by_component[
                component
            ].build_result._id

        return result

    def _reconcile_tree(
        self,
        builder: rio.Component,
        old_build_data: ComponentData,
        new_build: rio.Component,
    ) -> None:
        # Find all pairs of components which should be reconciled
        matched_pairs = list(
            self._find_components_for_reconciliation(
                old_build_data.build_result, new_build
            )
        )

        # Reconciliating individual components requires knowledge of which other
        # components are being reconciled.
        #
        # -> Collect them into a set first.
        reconciled_components_new_to_old: Dict[rio.Component, rio.Component] = {
            new_component: old_component
            for old_component, new_component in matched_pairs
        }

        # Reconcile all matched pairs
        for new_component, old_component in reconciled_components_new_to_old.items():
            self._reconcile_component(
                old_component,
                new_component,
                reconciled_components_new_to_old,
            )

            # Performance optimization: Since the new component has just been
            # reconciled into the old one, it cannot possibly still be part of
            # the component tree. It is thus safe to remove from the set of dirty
            # components to prevent a pointless rebuild.
            self._dirty_components.discard(new_component)

        # Update the component data. If the root component was not reconciled, the new
        # component is the new build result.
        try:
            reconciled_build_result = reconciled_components_new_to_old[new_build]
        except KeyError:
            reconciled_build_result = new_build
            old_build_data.build_result = new_build

        # Replace any references to new reconciled components to old ones instead
        def remap_components(parent: rio.Component) -> None:
            parent_vars = vars(parent)

            for attr_name in inspection.get_child_component_containing_attribute_names(
                type(parent)
            ):
                attr_value = parent_vars[attr_name]

                # Just a component
                if isinstance(attr_value, rio.Component):
                    try:
                        attr_value = reconciled_components_new_to_old[attr_value]
                    except KeyError:
                        # Make sure that any components which are now in the tree
                        # have their builder properly set.
                        if isinstance(parent, component_base.FundamentalComponent):
                            attr_value._weak_builder_ = parent._weak_builder_
                            attr_value._build_generation_ = parent._build_generation_
                    else:
                        parent_vars[attr_name] = attr_value

                    remap_components(attr_value)

                # List / Collection
                elif isinstance(attr_value, list):
                    attr_value = cast(List[object], attr_value)

                    for ii, item in enumerate(attr_value):
                        if isinstance(item, rio.Component):
                            try:
                                item = reconciled_components_new_to_old[item]
                            except KeyError:
                                # Make sure that any components which are now in
                                # the tree have their builder properly set.
                                if isinstance(
                                    parent, component_base.FundamentalComponent
                                ):
                                    item._weak_builder_ = parent._weak_builder_
                                    item._build_generation_ = parent._build_generation_
                            else:
                                attr_value[ii] = item

                            remap_components(item)

        remap_components(reconciled_build_result)

    def _reconcile_component(
        self,
        old_component: rio.Component,
        new_component: rio.Component,
        reconciled_components_new_to_old: Dict[rio.Component, rio.Component],
    ) -> None:
        """
        Given two components of the same type, reconcile them. Specifically:

        - Any state which was explicitly set by the user in the new component's
          constructor is considered explicitly set, and will be copied into the
          old component
        - If any values were changed, the component is registered as dirty with the
          session

        This function also handles `StateBinding`s, for details see comments.
        """
        assert type(old_component) is type(new_component), (
            old_component,
            new_component,
        )

        # Let any following code assume that the two components aren't the same
        # instance
        if old_component is new_component:
            return

        # Determine which properties will be taken from the new component
        overridden_values: Dict[str, object] = {}
        old_component_dict = vars(old_component)
        new_component_dict = vars(new_component)

        for prop_name in new_component._state_properties_:
            # Should the value be overridden?
            if prop_name not in new_component._explicitly_set_properties_:
                continue

            # Take care to keep state bindings up to date
            old_value = old_component_dict[prop_name]
            new_value = new_component_dict[prop_name]
            old_is_binding = isinstance(old_value, component_base.StateBinding)
            new_is_binding = isinstance(new_value, component_base.StateBinding)

            # If the old value was a binding, and the new one isn't, split the
            # tree of bindings. All children are now roots.
            if old_is_binding and not new_is_binding:
                binding_value = old_value.get_value()
                old_value.owning_component_weak = lambda: None

                for child_binding in old_value.children:
                    child_binding.is_root = True
                    child_binding.parent = None
                    child_binding.value = binding_value

            # If both values are bindings transfer the children to the new
            # binding
            elif old_is_binding and new_is_binding:
                new_value.owning_component_weak = old_value.owning_component_weak
                new_value.children = old_value.children

                for child in old_value.children:
                    child.parent = new_value

                # Save the binding's value in case this is the root binding
                new_value.value = old_value.value

            overridden_values[prop_name] = new_value

        # If the component has changed mark it as dirty
        def values_equal(old: object, new: object) -> bool:
            """
            Used to compare the old and new values of a property. Returns `True`
            if the values are considered equal, `False` otherwise.
            """
            # Components are a special case. Component attributes are dirty iff the
            # component isn't reconciled, i.e. it is a new component
            if isinstance(new, rio.Component):
                return old is new or old is reconciled_components_new_to_old.get(
                    new, None
                )

            if isinstance(new, list):
                if not isinstance(old, list):
                    return False

                old = cast(List[object], old)
                new = cast(List[object], new)

                if len(old) != len(new):
                    return False

                for old_item, new_item in zip(old, new):
                    if not values_equal(old_item, new_item):
                        return False

                return True

            # Otherwise attempt to compare the values
            try:
                return bool(old == new)
            except Exception:
                return old is new

        # Determine which properties will be taken from the new component
        for prop_name in overridden_values:
            old_value = getattr(old_component, prop_name)
            new_value = getattr(new_component, prop_name)

            if not values_equal(old_value, new_value):
                self._register_dirty_component(
                    old_component,
                    include_children_recursively=False,
                )
                break

        # Override the key now. It should never be preserved, but doesn't make
        # the component dirty
        overridden_values["key"] = new_component.key

        # Now combine the old and new dictionaries
        #
        # Notice that the component's `_weak_builder_` is always preserved. So even
        # components whose position in the tree has changed still have the correct
        # builder set.
        old_component_dict.update(overridden_values)

    def _find_components_for_reconciliation(
        self,
        old_build: rio.Component,
        new_build: rio.Component,
    ) -> Iterable[Tuple[rio.Component, rio.Component]]:
        """
        Given two component trees, find pairs of components which can be
        reconciled, i.e. which represent the "same" component. When exactly
        components are considered to be the same is up to the implementation and
        best-effort.

        Returns an iterable over (old_component, new_component) pairs, as well as a
        list of all components occurring in the new tree, which did not have a match
        in the old tree.
        """

        old_components_by_key: Dict[str, rio.Component] = {}
        new_components_by_key: Dict[str, rio.Component] = {}

        matches_by_topology: List[Tuple[rio.Component, rio.Component]] = []

        # First scan all components for topological matches, and also keep track of
        # each component by its key
        def register_component_by_key(
            components_by_key: Dict[str, rio.Component],
            component: rio.Component,
        ) -> None:
            if component.key is None:
                return

            # It's possible that the same component is registered twice, once
            # from a key_scan caused by a failed structural match, and once from
            # recursing into a successful key match.
            if (
                component.key in components_by_key
                and components_by_key[component.key] is not component
            ):
                raise RuntimeError(
                    f'Multiple components share the key "{component.key}": {components_by_key[component.key]} and {component}'
                )

            components_by_key[component.key] = component

        def key_scan(
            components_by_key: Dict[str, rio.Component],
            component: rio.Component,
            include_self: bool = True,
        ) -> None:
            for (
                child
            ) in component._iter_direct_and_indirect_child_containing_attributes(
                include_self=include_self,
                recurse_into_high_level_components=True,
            ):
                register_component_by_key(components_by_key, child)

        def chain_to_children(
            old_component: rio.Component,
            new_component: rio.Component,
        ) -> None:
            def _extract_components(attr: object) -> List[rio.Component]:
                if isinstance(attr, rio.Component):
                    return [attr]

                if isinstance(attr, list):
                    attr = cast(List[object], attr)

                    return [item for item in attr if isinstance(item, rio.Component)]

                return []

            # Iterate over the children, but make sure to preserve the topology.
            # Can't just use `iter_direct_children` here, since that would
            # discard topological information.
            for attr_name in inspection.get_child_component_containing_attribute_names(
                type(new_component)
            ):
                old_value = getattr(old_component, attr_name, None)
                new_value = getattr(new_component, attr_name, None)

                old_components = _extract_components(old_value)
                new_components = _extract_components(new_value)

                # Chain to the children
                common = min(len(old_components), len(new_components))
                for old_child, new_child in zip(old_components, new_components):
                    worker(old_child, new_child)

                for old_child in old_components[common:]:
                    key_scan(old_components_by_key, old_child, include_self=True)

                for new_child in new_components[common:]:
                    key_scan(new_components_by_key, new_child, include_self=True)

        def worker(old_component: rio.Component, new_component: rio.Component) -> None:
            # Register the component by key
            register_component_by_key(old_components_by_key, old_component)
            register_component_by_key(new_components_by_key, new_component)

            # If the components' types or keys don't match, stop looking for
            # topological matches. Just keep track of the children's keys.
            if (
                type(old_component) is not type(new_component)
                or old_component.key != new_component.key
            ):
                key_scan(old_components_by_key, old_component, include_self=False)
                key_scan(new_components_by_key, new_component, include_self=False)
                return

            # Key matches are handled elsewhere, so if the key is not `None`, do
            # nothing. We'd just end up doing the same work twice.
            if old_component.key is not None:
                return

            matches_by_topology.append((old_component, new_component))
            chain_to_children(old_component, new_component)

        worker(old_build, new_build)

        # Find matches by key and reconcile their children. This can produce new
        # key matches, so we do it in a loop.
        new_key_matches = old_components_by_key.keys() & new_components_by_key.keys()

        while new_key_matches:
            for key in new_key_matches:
                old_component = old_components_by_key[key]
                new_component = new_components_by_key[key]

                # If the components have different types, even the same key
                # can't make them match
                if type(old_component) is not type(new_component):
                    continue

                yield (old_component, new_component)

                # Recurse into these two components
                chain_to_children(old_component, new_component)

            # If any new key matches were found, repeat the process
            new_key_matches = (
                old_components_by_key.keys()
                & new_components_by_key.keys() - new_key_matches
            )

        # Yield topological matches
        for old_component, new_component in matches_by_topology:
            yield old_component, new_component

    def _register_font(self, font: text_style.Font) -> None:
        # Fonts are different from other assets because they need to be
        # registered under a name, not just a URL. We don't want to re-register
        # the same font multiple times, so we keep track of all registered
        # fonts. Every registered font is associated with all its assets
        # (regular, bold, italic, ...), which will be kept alive until the
        # session is closed.
        if font.name in self._registered_font_assets:
            return

        # It's a new font, register it
        font_assets: List[assets.Asset] = []
        urls: List[Optional[str]] = [None] * 4

        for i, location in enumerate(
            (font.regular, font.bold, font.italic, font.bold_italic)
        ):
            if location is None:
                continue

            # Host the font file as an asset
            asset = assets.Asset.new(location)
            urls[i] = asset._serialize(self)

            font_assets.append(asset)

        self.create_task(self._remote_register_font(font.name, urls))  # type: ignore

        self._registered_font_assets[font.name] = font_assets

    def _get_settings_file_path(self) -> Path:
        """
        The path to the settings file. Only used if `running_in_window`.
        """
        import platformdirs

        return (
            Path(
                platformdirs.user_data_dir(
                    appname=self._app_server.app.name,
                    roaming=True,
                    ensure_exists=True,
                )
            )
            / "settings.json"
        )

    async def _load_user_settings(
        self, settings_sent_by_client: Dict[str, object]
    ) -> None:
        # If `running_in_window`, load the settings from the config file.
        # Otherwise, parse the settings sent by the browser.
        #
        # Keys in this dict can be attributes of the "root" section or names of
        # sections. To prevent name clashes, section names are prefixed with
        # "section:".
        settings_json: Dict[str, object]

        if self.running_in_window:
            try:
                async with aiofiles.open(self._get_settings_file_path()) as file:
                    settings_text = await file.read()

                settings_json = json.loads(settings_text)
            except (IOError, json.JSONDecodeError):
                settings_json = {}

            self._settings_json = settings_json
        else:
            # Browsers send us a flat dict where the keys are prefixed with the
            # section name. We will convert each section into a dict.
            settings_json = {}

            for key, value in settings_sent_by_client.items():
                # Find the section name
                section_name, _, key = key.rpartition(":")

                if section_name:
                    section = settings_json.setdefault("section:" + section_name, {})
                    section[key] = value  # type: ignore
                else:
                    settings_json[key] = value

        # If `running_in_window`, we need to store the settings JSON so that
        # we can update the changed keys and write the whole thing to a file.
        if self.running_in_window:
            self._settings_json = settings_json

        # Instantiate and attach the settings
        for default_settings in self._app_server.default_attachments:
            if not isinstance(default_settings, user_settings_module.UserSettings):
                continue

            settings = type(default_settings)._from_json(
                self,
                settings_json,
                default_settings,
            )

            # Attach the instance to the session
            self.attachments._add(settings, synchronize=False)

    async def set_title(self, title: str) -> None:
        if self.running_in_window:
            window = await self._get_webview_window()
            window.set_title(title)
        else:
            await self._remote_set_title(title)

    @overload
    async def file_chooser(
        self,
        *,
        file_extensions: Optional[Iterable[str]] = None,
        multiple: Literal[False] = False,
    ) -> common.FileInfo:
        ...

    @overload
    async def file_chooser(
        self,
        *,
        file_extensions: Optional[Iterable[str]] = None,
        multiple: Literal[True],
    ) -> Tuple[common.FileInfo]:
        ...

    async def file_chooser(
        self,
        *,
        file_extensions: Optional[Iterable[str]] = None,
        multiple: bool = False,
    ) -> Union[common.FileInfo, Tuple[common.FileInfo]]:
        """
        Open a file chooser dialog.

        This function opens a file chooser dialog, allowing the user to select a
        file. The selected file is returned, allowing you to access its
        contents.

        See also `save_file`, if you want to save a file instead of opening one.

        Args:
            file_extensions: A list of file extensions which the user is allowed
                to select. Defaults to `None`, which means that the user may
                select any file.

            multiple: Whether the user should pick a single file, or multiple.

        Raises:
            NoFileSelectedError: If the user did not select a file.
        """
        # Create a secret id and register the file upload with the app server
        upload_id = secrets.token_urlsafe()
        future = asyncio.Future[List[common.FileInfo]]()

        self._app_server._pending_file_uploads[upload_id] = future

        # Allow the user to specify both `jpg` and `.jpg`
        if file_extensions is not None:
            file_extensions = [
                ext if ext.startswith(".") else f".{ext}" for ext in file_extensions
            ]

        # Tell the frontend to upload a file
        await self._request_file_upload(
            upload_url=f"/rio/upload/{upload_id}",
            file_extensions=file_extensions,
            multiple=multiple,
        )

        # Wait for the user to upload files
        files = await future

        # Raise an exception if no files were uploaded
        if not files:
            raise errors.NoFileSelectedError()

        # Ensure only one file was provided if `multiple` is False
        if not multiple and len(files) != 1:
            logging.warning(
                "Client attempted to upload multiple files when `multiple` was False."
            )
            raise errors.NoFileSelectedError()

        # Return the file info
        if multiple:
            return tuple(files)  # type: ignore
        else:
            return files[0]

    async def save_file(
        self,
        file_contents: Union[Path, str, bytes],
        file_name: str = "Unnamed File",
        *,
        media_type: Optional[str] = None,
        directory: Optional[Path] = None,
    ) -> None:
        """
        Save a file to the user's device.

        This function allows you to save a file to the user's device. The user
        will be prompted to select a location to save the file to.

        See also `file_chooser` if you want to open a file instead of saving
        one.

        Args:
            file_contents: The contents of the file to save. This can be a
                string, bytes, or a path to a file on the server.

            file_name: The default file name that will be displayed in the file
                dialog. The user can freely change it.

            media_type: The media type of the file. Defaults to `None`, which
                means that the media type will be guessed from the file name.

            directory: The directory where the file dialog should open. This has
                no effect if the user is visiting the app in a browser.
        """
        if self.running_in_window:
            # FIXME: Find (1) a better way to get the active window and (2) a
            # way to open a file dialog without blocking the event loop.
            import webview  # type: ignore

            window = await self._get_webview_window()
            destinations = window.create_file_dialog(
                webview.SAVE_DIALOG,
                directory="" if directory is None else str(directory),
                save_filename=file_name,
            )
            if not destinations:
                return  # TODO: Raise error?

            destination = destinations[0]

            if isinstance(file_contents, Path):
                await asyncio.to_thread(shutil.copy, file_contents, destination)

            elif isinstance(file_contents, str):
                async with aiofiles.open(destination, "w", encoding="utf8") as file:
                    await file.write(file_contents)

            elif isinstance(file_contents, bytes):  # type: ignore[unnecessary-isinstance]
                async with aiofiles.open(destination, "wb") as file:
                    await file.write(file_contents)

            else:
                raise ValueError(
                    f"The file contents must be a Path, str or bytes, not {file_contents!r}"
                )

            return

        # Create an asset for the file
        if isinstance(file_contents, Path):
            as_asset = assets.PathAsset(file_contents, media_type)

        elif isinstance(file_contents, str):
            as_asset = assets.BytesAsset(
                file_contents.encode("utf-8"),
                "text/plain" if media_type is None else media_type,
            )

        elif isinstance(file_contents, bytes):  # type: ignore[unnecessary-isinstance]
            as_asset = assets.BytesAsset(
                file_contents,
                "application/octet-stream" if media_type is None else media_type,
            )

        else:
            raise ValueError(
                f"The file contents must be a Path, str or bytes, not {file_contents!r}"
            )

        # Host the asset
        url = self._app_server.host_asset_with_timeout(as_asset, 60)

        # Tell the frontend to download the file
        await self._evaluate_javascript(
            f"""
const a = document.createElement('a')
a.href = {json.dumps(str(url))}
a.download = {json.dumps(file_name)}
a.type = {json.dumps(media_type)}
a.target = "_blank"
document.body.appendChild(a)
a.click()
document.body.removeChild(a)
"""
        )

    async def _apply_theme(self, thm: theme.Theme) -> None:
        """
        Updates the client's theme to match the given one.
        """
        self.theme = thm
        # Build the set of all CSS variables that must be set

        # Miscellaneous
        variables: Dict[str, str] = {
            "--rio-global-corner-radius-small": f"{thm.corner_radius_small}rem",
            "--rio-global-corner-radius-medium": f"{thm.corner_radius_medium}rem",
            "--rio-global-corner-radius-large": f"{thm.corner_radius_large}rem",
            "--rio-global-shadow-color": f"#{thm.shadow_color.hex}",
        }

        # Palettes
        palette_names = (
            "primary",
            "secondary",
            "background",
            "neutral",
            "disabled",
            "success",
            "warning",
            "danger",
        )

        for palette_name in palette_names:
            palette = getattr(thm, f"{palette_name}_palette")
            assert isinstance(palette, theme.Palette), palette

            variables[f"--rio-global-{palette_name}-bg"] = f"#{palette.background.hex}"
            variables[
                f"--rio-global-{palette_name}-bg-variant"
            ] = f"#{palette.background_variant.hex}"
            variables[
                f"--rio-global-{palette_name}-bg-active"
            ] = f"#{palette.background_active.hex}"
            variables[f"--rio-global-{palette_name}-fg"] = f"#{palette.foreground.hex}"

        # Text styles
        style_names = (
            "heading1",
            "heading2",
            "heading3",
            "text",
        )

        for style_name in style_names:
            style = getattr(thm, f"{style_name}_style")
            assert isinstance(style, rio.TextStyle), style

            css_prefix = f"--rio-global-{style_name}"
            variables[f"{css_prefix}-font-name"] = style.font._serialize(self)
            variables[f"{css_prefix}-font-size"] = f"{style.font_size}rem"
            variables[f"{css_prefix}-italic"] = "italic" if style.italic else "normal"
            variables[f"{css_prefix}-font-weight"] = style.font_weight
            variables[f"{css_prefix}-underlined"] = (
                "underline" if style.underlined else "unset"
            )
            variables[f"{css_prefix}-all-caps"] = (
                "uppercase" if style.all_caps else "unset"
            )

            # CSS variables for the fill
            fill_variables = _host_and_get_fill_as_css_variables(style.fill, self)

            for var, value in fill_variables.items():
                variables[f"{css_prefix}-{var}"] = value

        # Update the variables client-side
        await self._remote_apply_theme(
            variables,
            "light"
            if thm.neutral_palette.background.perceived_brightness > 0.5
            else "dark",
        )

    @unicall.remote(
        name="applyTheme",
        parameter_format="dict",
        await_response=False,
    )
    async def _remote_apply_theme(
        self,
        css_variables: Dict[str, str],
        theme_variant: Literal["light", "dark"],
    ) -> None:
        raise NotImplementedError

    @unicall.remote(
        name="setTitle",
        parameter_format="dict",
        await_response=False,
    )
    async def _remote_set_title(self, title: str) -> None:
        raise NotImplementedError

    @unicall.remote(
        name="setKeyboardFocus",
        parameter_format="dict",
        await_response=False,
    )
    async def _remote_set_keyboard_focus(self, component_id: int) -> None:
        raise NotImplementedError

    @unicall.remote(
        name="updateComponentStates",
        parameter_format="dict",
        await_response=False,
    )
    async def _remote_update_component_states(
        self,
        # Maps component ids to serialized components. The components may be partial,
        # i.e. any property may be missing.
        delta_states: Dict[int, Any],
        # Tells the client to make the given component the new root component.
        root_component_id: Optional[int],
    ) -> None:
        """
        Replace all components in the UI with the given one.
        """
        raise NotImplementedError

    @unicall.remote(
        name="evaluateJavaScript",
        parameter_format="dict",
        await_response=True,
    )
    async def _evaluate_javascript(self, java_script_source: str) -> Any:
        """
        Evaluate the given javascript code in the client.
        """
        raise NotImplementedError

    @unicall.remote(
        name="requestFileUpload",
        parameter_format="dict",
        await_response=False,
    )
    async def _request_file_upload(
        self,
        upload_url: str,
        file_extensions: Optional[List[str]],
        multiple: bool,
    ) -> None:
        """
        Tell the client to upload a file to the server.
        """
        raise NotImplementedError

    @unicall.remote(name="setUserSettings", await_response=False)
    async def _set_user_settings(self, delta_settings: Dict[str, Any]) -> None:
        """
        Persistently store the given key-value pairs at the user. The values
        have to be jsonable.

        Any keys not present here are still preserved. Thus the function
        effectively behaves like `dict.update`.
        """
        raise NotImplementedError

    @unicall.remote(name="registerFont", await_response=False)
    async def _remote_register_font(self, name: str, urls: List[Optional[str]]) -> None:
        raise NotImplementedError

    @unicall.remote(
        name="closeSession",
        await_response=False,
    )
    async def _remote_close_session(self) -> None:
        raise NotImplementedError

    def _try_get_component_for_message(
        self, component_id: int
    ) -> Optional[rio.Component]:
        """
        Attempts to get the component referenced by `component_id`. Returns `None` if
        there is no such component. This can happen during normal opration, e.g.
        because a component has been deleted while the message was in flight.
        """

        try:
            return self._weak_components_by_id[component_id]
        except KeyError:
            logging.warn(
                f"Encountered message for unknown component {component_id}. (The component might have been deleted in the meantime.)"
            )
            return None

    @unicall.local(name="componentStateUpdate")
    async def _component_state_update(
        self,
        component_id: int,
        delta_state: Any,
    ) -> None:
        # Get the component
        component = self._try_get_component_for_message(component_id)

        if component is None:
            return

        # Update the component's state
        assert isinstance(component, component_base.FundamentalComponent), component
        await component._on_state_update(delta_state)

    @unicall.local(name="componentMessage")
    async def _component_message(
        self,
        component_id: int,
        payload: Any,
    ) -> None:
        # Get the component
        component = self._try_get_component_for_message(component_id)

        if component is None:
            return

        # Let the component handle the message
        await component._on_message(payload)

    @unicall.local(name="ping")
    async def _ping(self, ping: str) -> str:
        return "pong"

    @unicall.local(name="onUrlChange")
    async def _on_url_change(self, new_url: str) -> None:
        """
        Called by the client when the page changes.
        """
        # Try to navigate to the new page
        self.navigate_to(
            new_url,
            replace=True,
        )

        # Refresh the session
        await self._refresh()

    @unicall.local(name="onWindowResize")
    async def _on_window_resize(self, new_width: float, new_height: float) -> None:
        """
        Called by the client when the window is resized.
        """
        # Update the stored window size
        self._window_width = new_width
        self._window_height = new_height

        # Call any registered callbacks
        for component, callback in self._on_window_resize_callbacks.items():
            self.create_task(
                self._call_event_handler(callback, component),
                name="`on_window_resize` event handler",
            )



from typing import *  # type: ignore
from typing import Optional

from .component_base import Component, FundamentalComponent

__all__ = ["ScrollTarget"]


class ScrollTarget(FundamentalComponent):
    """
    Allows browsers to scroll to a specific component via URL fragment.

    `ScrollTarget` is a container which can be referenced by a URL fragment,
    which allows browsers to scroll to it when the page is loaded. For example,
    if your website contains a `ScrollTarget` with the ID `"my-section"`, then a
    browser visiting `https://your.website/#my-section` will immediately scroll
    it into view.

    Attributes:
        id: The ID of the `ScrollTarget`. This must be unique among all
            `ScrollTarget`s on the page.

        child: The child component to display inside the `ScrollTarget`.
    """

    id: str
    child: Optional[Component] = None

    def __post_init__(self):
        if self.id.startswith("rio-id-"):
            raise ValueError(
                f"Invalid ID for ScrollTarget: `{self.id}`. IDs starting with"
                f"`rio-id-` are reserved for internal use."
            )


ScrollTarget._unique_id = "ScrollTarget-builtin"



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import *  # type: ignore

from . import component_base

__all__ = [
    "ProgressBar",
]


class ProgressBar(component_base.FundamentalComponent):
    """
    A progress indicator in the shape of a horizontal bar.

    `ProgressBar` conveys to the user that activity is ongoing. It can either
    display the exact progress as a fraction from 0 to 1, or it can display an
    indeterminate progress animation, which is useful when the exact progress
    isn't known.

    The circular counterpart to this component is the `ProgressCircle`.

    Attributes:
        progress: The progress to display, as a fraction from 0 to 1. If `None`,
            the progress indicator will be indeterminate.
    """

    progress: Optional[float] = None


ProgressBar._unique_id = "ProgressBar-builtin"



from __future__ import annotations

import enum
from dataclasses import KW_ONLY
from typing import *  # type: ignore

import rio

from .. import inspection
from . import banner, component_base, number_input, switch, text

__all__ = ["AutoForm"]


def prettify_name(name: str) -> str:
    parts = name.split("_")
    return " ".join(p.title() for p in parts)


class AutoForm(component_base.Component):
    # Not yet ready for release - do not use.

    _: KW_ONLY
    spacing: float = 0.5
    label_width: float = 15.0
    error_message: Optional[str] = None
    submit_text: str = "Submit"

    # Internal
    _is_loading: bool = False

    def validate(self) -> Optional[str]:
        """
        Override this method to validate the form's values. If a string is
        returned it will be displayed to the user, and the form will not be
        submittable.
        """
        return None

    async def on_submit(self) -> None:
        """
        Called when the user clicks the submit button. Override this method to
        perform some action when the form is submitted.
        """
        pass

    async def _on_submit(self) -> None:
        # Display the loading indicator
        self._is_loading = True
        await self.force_refresh()

        # Delegate to the user's submit method
        try:
            await self.call_event_handler(self.on_submit)
        finally:
            self._is_loading = False

    def make_notification_bar(
        self,
        text: str,
        level: Literal["info", "warning", "error"],
    ) -> rio.Component:
        return banner.Banner(
            text=text,
            style=level,
        )

    def make_switch(self, is_on: bool) -> rio.Component:
        return switch.Switch(is_on=is_on)

    def make_button(
        self,
        *,
        text: str,
        on_press: rio.EventHandler[[]],
        is_sensitive: bool,
        is_loading: bool,
        is_major: bool,
    ) -> rio.Component:
        return rio.Button(
            text_or_child == text,
            style="major" if is_major else "minor",
            on_press=on_press,
            is_sensitive=is_sensitive,
            is_loading=is_loading,
        )

    def _build_input_field(
        self,
        field_name: str,
        field_type: type,
    ) -> rio.Component:
        # Get sensible type information
        origin = get_origin(field_type)
        field_args = get_args(field_type)
        field_type = field_type if origin is None else origin
        del origin

        # Grab the field instance
        field_instance = vars(type(self))[field_name]

        # `bool` -> `Switch`
        if field_type is bool:
            return self.make_switch(
                is_on=field_instance,
            )

        # `int` -> `NumberInput`
        if field_type is int:
            return number_input.NumberInput(
                value=field_instance,
                decimals=0,
            )

        # `float` -> `NumberInput`
        if field_type is float:
            return number_input.NumberInput(
                value=field_instance,
            )

        # `str` -> `TextInput`
        if field_type is str:
            return rio.TextInput(
                text=field_instance,
            )

        # `Literal` or `Enum` -> `Dropdown`
        if field_type is Literal or issubclass(field_type, enum.Enum):
            if field_type is Literal:
                mapping = {str(a): a for a in field_args}
            else:
                mapping = {prettify_name(f.name): f.value for f in field_type}

            return rio.Dropdown(
                mapping,
                selected_value=field_instance,
            )

        # Unsupported type
        raise TypeError(f"AutoForm does not support fields of type `{field_type}`")

    def build(self) -> rio.Component:
        rows: List[rio.Component] = []

        # Look for errors
        error_message = (
            self.validate() if self.error_message is None else self.error_message
        )

        # Display any errors
        if error_message is not None:
            rows.append(
                self.make_notification_bar(
                    text=error_message,
                    level="error",
                )
            )

        # One row per field
        for field_name, field_type in inspection.get_type_annotations(
            type(self)
        ).items():
            rows.append(
                rio.Row(
                    text.Text(
                        prettify_name(field_name),
                        width=self.label_width,
                    ),
                    self._build_input_field(field_name, field_type),
                    spacing=self.spacing,
                )
            )

        # Add a submit button
        rows.append(
            self.make_button(
                text=self.submit_text,
                on_press=self._on_submit,
                is_sensitive=not self._is_loading and error_message is None,
                is_loading=self._is_loading,
                is_major=True,
            )
        )

        # Wrap everything in one container
        return rio.Column(
            *rows,
            spacing=self.spacing,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import *  # type: ignore

import uniserde
from uniserde import Jsonable

import rio
import rio.inspection
from rio.components.component_base import FundamentalComponent

DUMP_PATH = rio.common.PROJECT_ROOT_DIR / "tree-dump.json"


__all__ = [
    "Attribute",
    "Component",
    "dump_tree",
    "TreeDump",
]


@uniserde.as_child
@dataclass
class Attribute:
    name: str
    type: str
    value: str | List[int]
    bound_to: Optional[Tuple[str, str]]


@dataclass
class Component(uniserde.Serde):
    id: int
    type: str
    parent: Optional[int]
    builder: Optional[int]
    attributes: List[Attribute]


@dataclass
class TreeDump(uniserde.Serde):
    components: List[Component]


# def _get_attr_value(attr_value: Any) -> Union[str, List[int]]:
#     if prop.name in child_containing_attribute_names:
#         if attr_value is None:
#             attr_children = []
#         elif isinstance(attr_value, rio.Component):
#             attr_children = [attr_value]
#         elif isinstance(attr_value, (list, tuple)):
#             attr_children = attr_value
#         elif isinstance(attr_value, dict):
#             attr_children = attr_value.values()
#         else:
#             raise NotImplementedError(
#                 f"Unsupported type for attribute {prop.name}: {type(attr_value)}"
#             )

#         attr_value = [child._id for child in attr_children]
#     else:
#         attr_value = str(attr_value)


def _dump_worker(
    sess: rio.Session,
    component: rio.Component,
    parent: Optional[int],
    builder: Optional[int],
) -> Iterable[Component]:
    # Fetch all of the component's attributes
    attributes: List[Attribute] = []
    # child_containing_attribute_names = (
    #     rio.inspection.get_child_component_containing_attribute_names(type(component))
    # )

    for prop in type(component)._state_properties_.values():
        raw_attr_value = prop.__get__(component, None)

        attributes.append(
            Attribute(
                name=prop.name,
                type="TODO-type",
                # value=_get_attr_value(raw_attr_value),
                value=str(raw_attr_value),
                bound_to=None,  # TODO
            )
        )

    # For fundamental components, all children are in the tree and need to be
    # processed. For non-fundamental components, fetch the build result
    if isinstance(component, FundamentalComponent):
        children = list(component._iter_direct_children())
        child_builder = builder
    else:
        component_data = sess._weak_component_data_by_component[component]
        children = [component_data.build_result]
        child_builder = component._id

    # Process all children recursively
    for child in children:
        yield from _dump_worker(
            sess=sess,
            component=child,
            parent=component._id,
            builder=child_builder,
        )

    # Instantiate the result
    yield Component(
        id=component._id,
        type=type(component).__name__,
        parent=parent,
        builder=builder,
        attributes=attributes,
    )


def dump_tree(sess: rio.Session) -> TreeDump:
    # Make sure there are no pending changes
    sess._refresh_sync()

    # Dump all components
    components = list(
        _dump_worker(
            sess=sess,
            component=sess._root_component,
            parent=None,
            builder=None,
        )
    )

    # Instantiate the result
    return TreeDump(
        components=components,
    )



import rio


# <code>
class CustomComponent(rio.Component):
    some_value: str

    def build(self) -> rio.Component:
        return rio.Column(
            rio.TextInput(
                text=CustomComponent.some_value,
            ),
            rio.TextInput(
                text=CustomComponent.some_value,
            ),
            spacing=1,
        )


# </code>



from __future__ import annotations

from abc import ABC, abstractmethod

from uniserde import Jsonable

import rio

__all__ = ["SelfSerializing"]


class SelfSerializing(ABC):
    """
    Properties with types that inherit from `SelfSerializing` will be serialized
    when sent to the client.
    """

    @abstractmethod
    def _serialize(self, sess: rio.Session) -> Jsonable:
        raise NotImplementedError



import rio


# <code>
class CustomComponent(rio.Component):
    some_value: str

    def _on_button_press(self) -> None:
        print(
            self.some_value
        )  # Just read the local value. It will always be up-to-date

    def build(self) -> rio.Component:
        return rio.Column(
            rio.TextInput(
                text=CustomComponent.some_value,  # Using the class property creates a state binding
            ),
            rio.Button(
                "Click me!",
                on_press=self._on_button_press,
            ),
        )


# </code>



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import *  # type: ignore

from uniserde import JsonDoc

from .. import assets
from ..common import EventHandler, ImageLike
from . import component_base

__all__ = ["Image"]


class Image(component_base.FundamentalComponent):
    """
    Displays an image file.

    `Image` does just what you'd expect: it displays a single image. The image
    can be loaded from a URL or a local file.

    Note that the resolution of the image does not affect the size at which it
    is displayed. The `Image` component doesn't reserve any space for itself, it
    simply makes do with the space it is given by its parent component.

    The image can be scaled to fit the shape in one of three ways:

    - `fit`: The image is scaled to fit entirely inside the shape, while
      maintaining its aspect ratio. This is the default.
    - `stretch`: The image is stretched to fill the shape, distorting it if
      necessary.
    - `zoom`: The image is scaled to fill the shape entirely, while maintaining
      its aspect ratio. This may cause the image to overflow the shape.


    Attributes:
        image: The image to display.

        fill_mode: How the image should be scaled to fit the shape. If `fit`,
            the image is scaled to fit entirely inside the shape. If `stretch`,
            the image is stretched to fill the shape exactly, possibly
            distorting it in the process. If `zoom`, the image is scaled to fill
            the shape entirely, possibly overflowing.

        on_error: Triggered when the image fails to load.

        corner_radius: How round to make the corners of the image. If a single
            number is given, all four corners will be rounded equally. If a
            tuple of four numbers is given, they will be interpreted as the
            radii of the top-left, top-right, bottom-right, and bottom-left
            corners, in that order.
    """

    image: ImageLike
    _: KW_ONLY
    fill_mode: Literal["fit", "stretch", "zoom"] = "fit"
    on_error: EventHandler[[]] = None
    corner_radius: Union[float, Tuple[float, float, float, float]] = 0

    def _get_image_asset(self) -> assets.Asset:
        image = self.image

        if getattr(self, "_image_for_cached_asset", None) != image:
            self._cached_image_asset = assets.Asset.from_image(image)
            self._image_for_cached_asset = image

        return self._cached_image_asset

    def _custom_serialize(self) -> JsonDoc:
        if isinstance(self.corner_radius, (int, float)):
            corner_radius = (self.corner_radius,) * 4
        else:
            corner_radius = self.corner_radius

        return {
            "imageUrl": self._get_image_asset()._serialize(self.session),
            "reportError": self.on_error is not None,
            "corner_radius": corner_radius,
        }

    async def _on_message(self, message: JsonDoc) -> None:
        await self.call_event_handler(self.on_error)


Image._unique_id = "Image-builtin"



from dataclasses import KW_ONLY
from typing import *  # type: ignore

from .component_base import Component, FundamentalComponent

__all__ = ["ScrollContainer"]


class ScrollContainer(FundamentalComponent):
    """
    Displays a scroll bar if its child grows too large.

    `ScrollContainer` is a container which displays a scroll bar if its child
    component grows too large. It can scroll vertically and/or horizontally.

    Attributes:
        child: The child component to display inside the `ScrollContainer`.

        scroll_x: Controls horizontal scrolling. The default is `"auto"`, which
            means that a scroll bar will be displayed only if it is needed.
            `"always"` displays a scroll bar even if it isn't needed, and
            `"never"` disables horizontal scrolling altogether.

        scroll_y: Controls vertical scrolling. The default is `"auto"`, which
            means that a scroll bar will be displayed only if it is needed.
            `"always"` displays a scroll bar even if it isn't needed, and
            `"never"` disables vertical scrolling altogether.
    """

    child: Component
    _: KW_ONLY
    scroll_x: Literal["never", "auto", "always"] = "auto"
    scroll_y: Literal["never", "auto", "always"] = "auto"


ScrollContainer._unique_id = "ScrollContainer-builtin"



from __future__ import annotations

from dataclasses import dataclass
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from . import component_base

__all__ = [
    "Slider",
    "SliderChangeEvent",
]


T = TypeVar("T", bound=float)


@dataclass
class SliderChangeEvent(Generic[T]):
    value: T


class Slider(component_base.FundamentalComponent, Generic[T]):
    """
    A component for selecting a single value from a range.

    The `Slider` components allows the user to select a single real number value
    by dragging a handle along a line. The value can be any number within a
    range you can specify.

    Attributes:
        min: The minimum value the slider can be set to.

        max: The maximum value the slider can be set to.

        value: The current value of the slider.

        is_sensitive: Whether the slider should respond to user input.
    """

    min: T
    max: T
    value: T
    discrete: bool
    is_sensitive: bool
    on_change: rio.EventHandler[SliderChangeEvent[T]]

    @overload
    def __new__(
        cls,
        *,
        min: int = 0,
        max: int = 100,
        value: Optional[int] = None,
        discrete: Literal[True] = True,
        is_sensitive: bool = True,
        on_change: rio.EventHandler[SliderChangeEvent[int]] = None,
        key: Optional[str] = None,
        margin: Optional[float] = None,
        margin_x: Optional[float] = None,
        margin_y: Optional[float] = None,
        margin_left: Optional[float] = None,
        margin_top: Optional[float] = None,
        margin_right: Optional[float] = None,
        margin_bottom: Optional[float] = None,
        width: Union[Literal["grow"], float] = 1.3,
        height: Union[Literal["grow"], float] = 1.3,
        align_x: Optional[float] = None,
        align_y: Optional[float] = None,
    ) -> Slider[int]:
        ...

    @overload
    def __new__(
        cls,
        *,
        min: float = 0,
        max: float = 1,
        value: Optional[float] = None,
        discrete: bool = False,
        is_sensitive: bool = True,
        on_change: rio.EventHandler[SliderChangeEvent[float]] = None,
        key: Optional[str] = None,
        margin: Optional[float] = None,
        margin_x: Optional[float] = None,
        margin_y: Optional[float] = None,
        margin_left: Optional[float] = None,
        margin_top: Optional[float] = None,
        margin_right: Optional[float] = None,
        margin_bottom: Optional[float] = None,
        width: Union[Literal["grow"], float] = 1.3,
        height: Union[Literal["grow"], float] = 1.3,
        align_x: Optional[float] = None,
        align_y: Optional[float] = None,
    ) -> Slider[float]:
        ...

    def __new__(cls, *args, **kwargs):  # type: ignore
        return super().__new__(cls)

    def __init__(
        self,
        *,
        min: Optional[float] = None,
        max: Optional[float] = None,
        value: Optional[float] = None,
        discrete: bool = False,
        is_sensitive: bool = True,
        on_change: rio.EventHandler[SliderChangeEvent[T]] = None,
        key: Optional[str] = None,
        margin: Optional[float] = None,
        margin_x: Optional[float] = None,
        margin_y: Optional[float] = None,
        margin_left: Optional[float] = None,
        margin_top: Optional[float] = None,
        margin_right: Optional[float] = None,
        margin_bottom: Optional[float] = None,
        width: Union[Literal["grow"], float] = 1.3,
        height: Union[Literal["grow"], float] = 1.3,
        align_x: Optional[float] = None,
        align_y: Optional[float] = None,
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        if min is None:
            min = 0

        if max is None:
            max = 100 if discrete else 1

        if value is None:
            value = min + (max - min) / 2

            if discrete:
                value = round(value)

        self.min = min  # type: ignore
        self.max = max  # type: ignore
        self.value = value  # type: ignore
        self.discrete = discrete
        self.is_sensitive = is_sensitive
        self.on_change = on_change

    # TODO: When `min` or `max` is changed, make sure the value is still within
    # the range

    def _custom_serialize(self) -> JsonDoc:
        return {
            "min": self.min,
            "max": self.max,
            "value": self.value,
        }

    async def _on_state_update(self, delta_state: JsonDoc) -> None:
        # Trigger on_change event
        try:
            new_value = delta_state["value"]
        except KeyError:
            pass
        else:
            assert isinstance(new_value, (int, float)), new_value
            await self.call_event_handler(
                self.on_change,  # type: ignore
                SliderChangeEvent(new_value),
            )

        # Chain up
        await super()._on_state_update(delta_state)


Slider._unique_id = "Slider-builtin"



from __future__ import annotations

from dataclasses import KW_ONLY
from pathlib import Path
from typing import *  # type: ignore

from uniserde import JsonDoc
from yarl import URL

from .. import assets
from ..common import URL, EventHandler
from . import component_base

__all__ = ["MediaPlayer"]


class MediaPlayer(component_base.KeyboardFocusableFundamentalComponent):
    """
    Plays audio and video.

    `MediaPlayer` plays back audio and video files. It can play local files and
    URLs.

    Note that the `MediaPlayer` component doesn't reserve any space for itself,
    it simply makes do with the space it is given by its parent component.

    Attributes:
        media: The media to play. This can be a file path, URL, or bytes.

        media_type: The mime type of the media file. May help the browser to
            play the file correctly.

        loop: Whether to automatically restart from the beginning when the
            playback ends.

        autoplay: Whether to start playing the media automatically, without
            requiring the user to press "Play".

        controls: Whether to display controls like a Play/Pause button, volume
            slider, etc.

        muted: Whether to mute the audio.

        volume: The volume to play the audio at. 1.0 is the native volume;
            larger numbers increase the volume, smaller numbers decrease it.

        on_playback_end: An event handler to call when the media finishes
            playing.

        on_error: An event handler to call when an error occurs, for example if
            the file format isn't supported.
    """

    media: Union[Path, URL, bytes]
    media_type: Optional[str] = None
    _: KW_ONLY
    loop: bool = False
    autoplay: bool = False
    controls: bool = True
    muted: bool = False
    volume: float = 1.0
    on_playback_end: EventHandler[[]] = None
    on_error: EventHandler[[]] = None

    def _custom_serialize(self) -> JsonDoc:
        media_asset = assets.Asset.new(self.media, self.media_type)
        return {
            "mediaUrl": media_asset._serialize(self.session),
            "reportError": self.on_error is not None,
            "reportPlaybackEnd": self.on_playback_end is not None,
        }

    async def _on_message(self, message: JsonDoc) -> None:
        if message["type"] == "playbackEnd":
            await self.call_event_handler(self.on_playback_end)
        elif message["type"] == "error":
            await self.call_event_handler(self.on_error)


MediaPlayer._unique_id = "MediaPlayer-builtin"



import os
import shlex
from typing import *  # type: ignore

_ACTIVE_SHELL: Optional[Literal["bourne", "cmd.exe", "powershell"]] = None


def _guess_shell_non_cached() -> Literal["bourne", "cmd.exe", "powershell"]:
    # If the SHELL environment variable is set, use that
    try:
        shell_var = os.environ["SHELL"]
    except KeyError:
        pass
    else:
        if (
            shell_var.endswith("bash")
            or shell_var.endswith("zsh")
            or shell_var.endswith("sh")
        ):
            return "bourne"

        # TODO: Does windows actually expose the shell in an environment variable?
        if shell_var.endswith("cmd.exe"):
            return "cmd.exe"

        # TODO: Does windows actually expose the shell in an environment variable?
        if shell_var.endswith("powershell"):
            return "powershell"

    # Fall back to guessing from the operating system
    if os.name == "nt":
        return "cmd.exe"

    return "bourne"


def guess_shell() -> Literal["bourne", "cmd.exe", "powershell"]:
    """
    Attempt to determine the shell this program is running in.
    """
    global _ACTIVE_SHELL

    if _ACTIVE_SHELL is None:
        _ACTIVE_SHELL = _guess_shell_non_cached()

    return _ACTIVE_SHELL


def shell_escape(value: Any) -> str:
    """
    Given any value, return a string that represents the same value in a shell.
    If the value is not a string it will be converted to a string using `str`.

    Warning: This function is not guaranteed to be safe. It is a rather
        simplistic implementation that will _usually_ work.
    """
    value = str(value)

    # Try to determine the shell
    shell = guess_shell()

    # Escape the value, based on the shell
    if shell == "bourne":
        return shlex.quote(value)

    if shell == "cmd.exe":
        value = value.replace('"', '""')
        return f'"{value}"'

    if shell == "powershell":
        value = value.replace("'", "''")
        return f"'{value}'"

    raise NotImplementedError(f"Unknown/Invalid shell `{shell}`")



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import *  # type: ignore

from uniserde import JsonDoc

from .. import text_style
from . import component_base

__all__ = [
    "Text",
]


class Text(component_base.FundamentalComponent):
    """
    Displays unfornatted text.

    `Text` displays text without any formatting, making it one of the most
    commonly used components in Rio.

    While the text itself is unformatted, you can still control the style of
    the text using the `style` attribute. This allows you to change the font
    size, color, and more.

    Attributes:
        text: The text to display.

        multiline: Whether the text may be split into multiple lines if not
            enough space is available.

        selectable: Whether the text can be selected by the user.

        style: The style of the text. This can either be a `TextStyle` instance,
            or one of the built-in styles: `heading1`, `heading2`, `heading3`,
            or `text`.
    """

    text: str
    _: KW_ONLY
    multiline: bool = False
    selectable: bool = False
    style: Union[
        Literal["heading1", "heading2", "heading3", "text", "dim"],
        text_style.TextStyle,
    ] = "text"

    def _custom_serialize(self) -> JsonDoc:
        # Serialization doesn't handle unions. Hence the custom serialization
        # here
        if isinstance(self.style, str):
            style = self.style
        else:
            style = self.session._serialize_and_host_value(
                self.style,
                text_style.TextStyle,
            )

        return {
            "style": style,
        }


Text._unique_id = "Text-builtin"



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import *  # type: ignore

import rio

from .. import cursor_style
from . import component_base

__all__ = [
    "Rectangle",
]


class Rectangle(component_base.FundamentalComponent):
    """
    A customizable rectangle shape.

    Rectangles are versatile components that can be used to build up more
    complex elements. While not particularly interesting on their own, combining
    a rectangle with other components allows you to quickly create custom
    buttons, cards, or anything els you may need in your app.

    Rectangles also act as a simple source of animations. They accept two
    styles: A default style for when the user isn't interacting with them, and a
    hover style for when the mouse hovers above them. This, along with their
    `transition_time` attribute allows you to make your app feel dynamic and
    alive.

    Attributes:
        style: How the rectangle should appear when the user isn't interacting
            with it.

        child: The component to display inside the rectangle.

        hover_style: The style of the rectangle when the mouse hovers above it.
            If set to `None`, the rectangle will not change its appearance when
            hovered.

        transition_time: How many seconds it should take for the rectangle to
            transition between its styles.

        cursor: The cursor to display when the mouse hovers above the rectangle.

        ripple: Whether to display a Material Design ripple effect when the
            rectangle is hovered or clicked.
    """

    _: KW_ONLY
    style: rio.BoxStyle
    child: Optional[rio.Component] = None
    hover_style: Optional[rio.BoxStyle] = None
    transition_time: float = 1.0
    cursor: rio.CursorStyle = cursor_style.CursorStyle.DEFAULT
    ripple: bool = False


Rectangle._unique_id = "Rectangle-builtin"



from typing import *

from uniserde import JsonDoc  # type: ignore

from .component_base import FundamentalComponent

__all__ = ["Table"]


TableValue = Union[int, float, str]


class Table(FundamentalComponent):
    data: Union[Mapping[str, Sequence[TableValue]], Sequence[Sequence[TableValue]]]
    show_row_numbers: bool = True

    def _custom_serialize(self) -> JsonDoc:
        data = self.data

        if isinstance(data, list):
            data = {header: column for header, *column in zip(*data)}

        return {
            "data": data,  # type: ignore
        }


Table._unique_id = "Table-builtin"



from __future__ import annotations

from typing import Callable

import rio

from .component_base import Component, FundamentalComponent


# The HighLevelRootComponent is the top-most element. It's a high-level element
# that is never re-built, which is convenient in many ways:
# 1. Every component except for the root component itself has a valid builder
# 2. The JS code is simpler because the root component can't have an
#    alignment or margin
# 3. Children of non-fundamental components are automatically initialized
#    correctly, so we don't need to duplicate that logic here
class HighLevelRootComponent(Component):
    build_function: Callable[[], Component]
    build_connection_lost_message_function: Callable[[], Component]

    def build(self) -> Component:
        return FundamentalRootComponent(
            self.build_function(),
            self.build_connection_lost_message_function(),
        )


class FundamentalRootComponent(FundamentalComponent):
    child: Component
    connection_lost_component: Component


FundamentalRootComponent._unique_id = "FundamentalRootComponent-builtin"



import rio


# <circular-button>
class CircularButton(rio.Component):
    text: str
    on_press: rio.EventHandler[[]] = None

    def build(self):
        return rio.Button(
            self.text,
            on_press=self.on_press,
            shape="circle",
            width=4,
            height=4,
        )


# </circular-button>


# <counter>
class Counter(rio.Component):
    value: int = 0

    def _on_increment(self) -> None:
        self.value += 1

    def _on_decrement(self) -> None:
        self.value -= 1

    def _on_reset(self) -> None:
        self.value = 0

    def build(self):
        return rio.Column(
            rio.Row(
                CircularButton(
                    "-1",
                    on_press=self._on_decrement,
                ),
                rio.Text(
                    str(self.value),
                    style="heading1",
                    width=5,
                ),
                CircularButton(
                    "+1",
                    on_press=self._on_increment,
                ),
                spacing=3,
            ),
            rio.Button(
                "Reset",
                style="minor",
                on_press=self._on_reset,
            ),
            spacing=2,
            align_x=0.5,
            align_y=0.5,
        )


# </counter>


# <app>
app = rio.App(
    pages=[
        rio.Page("", Counter),
    ],
)
# </app>

# <run>
app.run_in_browser()
# </run>



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from . import component_base

__all__ = [
    "Switch",
    "SwitchChangeEvent",
]


@dataclass
class SwitchChangeEvent:
    is_on: bool


class Switch(component_base.FundamentalComponent):
    """
    An input for `True` / `False` values.

    Switches allow the user to toggle between an "on" and an "off" state. They
    thus correspond to a Python `bool` value. Use them to allow the user to
    enable or disable certain features, or to select between two options.

    Attributes:
        is_on: Whether the switch is currently in the "on" state.

        is_sensitive: Whether the switch should respond to user input.

        on_change: Triggered when the user toggles the switch.
    """

    is_on: bool = False
    _: KW_ONLY
    is_sensitive: bool = True
    on_change: rio.EventHandler[SwitchChangeEvent] = None

    async def _on_state_update(self, delta_state: JsonDoc) -> None:
        # Trigger on_change event
        try:
            new_value = delta_state["is_on"]
        except KeyError:
            pass
        else:
            assert isinstance(new_value, bool), new_value
            await self.call_event_handler(
                self.on_change,
                SwitchChangeEvent(new_value),
            )

        # Chain up
        await super()._on_state_update(delta_state)


Switch._unique_id = "Switch-builtin"



from __future__ import annotations

import copy
from dataclasses import KW_ONLY
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from . import component_base

try:
    import plotly  # type: ignore
    import plotly.graph_objects  # type: ignore
except ImportError:
    if TYPE_CHECKING:
        import plotly  # type: ignore


__all__ = [
    "Plot",
]


class Plot(component_base.FundamentalComponent):
    """
    Displays a graph.


    Attributes:
        figure: The plotly figure to display.

        style: Controls the appearance of the plot. The following attributes are supported:
            - `fill`
            - `corner_radius`
    """

    figure: plotly.graph_objects.Figure
    _: KW_ONLY
    style: Optional[rio.BoxStyle] = None

    def _custom_serialize(self) -> JsonDoc:
        # Determine a style
        if self.style is None:
            thm = self.session.theme
            box_style = rio.BoxStyle(
                fill=thm.neutral_palette.background_variant,
                corner_radius=thm.corner_radius_small,
            )
        else:
            box_style = self.style

        # Make the plot transparent, so the background configured by JavaScript
        # shines through.
        figure = copy.copy(self.figure)
        figure.update_layout(
            {
                "plot_bgcolor": "rgba(0,0,0,0)",
                "paper_bgcolor": "rgba(0,0,0,0)",
            }
        )

        return {
            "plotJson": figure.to_json(),
            "boxStyle": box_style._serialize(self.session),
        }


Plot._unique_id = "Plot-builtin"



from pathlib import Path
from typing import *  # type: ignore

import revel
import uvicorn
from revel import error, fatal, print, success, warning
from typing_extensions import TypeAlias

import rio.snippets

from . import project, project_setup

__all__ = [
    "app",
]

revel.GLOBAL_STYLES.add_alias("primary", ["magenta"])
revel.GLOBAL_STYLES.add_alias("bg-primary", ["bg-magenta"])

app = revel.App(
    summary="An easy to use, app & web framework for Python",
    details="""
Rio is a framework for building reactive apps and websites in Python. It's
designed to be easy to use, and to get out of your way as much as possible.

This is the command line interface for Rio. You can use it to easily create new
projects, run them, and more.
""",
)


@app.command(
    aliases={"init", "create"},
    summary="Create a new Rio project",
    details="""
The `new` command creates a new directory and populates it with the files needed
to start a new Rio project. You can optionally specify a template, in which case
the files from that template will be copied into the new project, allowing you
hit the ground running.
""",
    parameters=[
        revel.Parameter(
            "nicename",
            summary="Human-readable name for the new project",
            prompt="What should the project be called?",
        ),
        revel.Parameter(
            "type",
            summary="Whether to create a website or an app",
        ),
        revel.Parameter(
            "template",
            summary="Template to use for the new project",
        ),
    ],
)
def new(
    nicename: str,
    *,
    type: Literal["app", "website"],
    template: Optional[project_setup.TemplatesLiteral] = None,
) -> None:
    # Project setup is surprisingly complex. Handle it in another file
    project_setup.create_project(
        nicename=nicename,
        type=type,
        template=template,
    )


@app.command(
    summary="Run the current project",
    parameters=[
        revel.Parameter(
            "port",
            summary="Port to run the HTTP server on",
        ),
        revel.Parameter(
            "public",
            summary="Whether the app should be available on the local network, or just the local device",
        ),
    ],
)
def run(
    *,
    port: Optional[int] = None,
    public: bool = False,
) -> None:
    # TODO: Verify the parameters are okay. i.e. `port` is a valid port number

    # Load the project
    with project.RioProject.try_load() as proj:
        main_module = proj.main_module
        fastapi_app_variable = proj.fastapi_app_variable

    # Delegate to uvicorn
    uvicorn.run(
        f"{main_module}:{fastapi_app_variable}",
        host="0.0.0.0" if public else "127.0.0.1",
        port=proj.debug_port if port is None else port,
        reload=True,
    )



from typing import *  # type: ignore

import rio


class Project(rio.Component):
    name: str
    details: str
    link: rio.URL

    def on_press(self) -> None:
        self.session.navigate_to(self.link)

    def build(self) -> rio.Component:
        return rio.Card(
            rio.Column(
                rio.Text(
                    self.name,
                    style="heading3",
                ),
                rio.Text(
                    self.details,
                    multiline=True,
                ),
                spacing=0.5,
            ),
            on_press=self.on_press,
        )


class Projects(rio.Component):
    def build(self) -> rio.Component:
        return rio.Grid(
            (
                Project(
                    "uniserde",
                    "Pure Python, easy to use Serializer",
                    rio.URL("https://gitlab.com/Vivern/uniserde"),
                ),
                Project(
                    "uniserde",
                    "Pure Python, easy to use Serializer",
                    rio.URL("https://gitlab.com/Vivern/uniserde"),
                ),
                Project(
                    "uniserde",
                    "Pure Python, easy to use Serializer",
                    rio.URL("https://gitlab.com/Vivern/uniserde"),
                ),
            ),
            (
                Project(
                    "uniserde",
                    "Pure Python, easy to use Serializer",
                    rio.URL("https://gitlab.com/Vivern/uniserde"),
                ),
                Project(
                    "uniserde",
                    "Pure Python, easy to use Serializer",
                    rio.URL("https://gitlab.com/Vivern/uniserde"),
                ),
                Project(
                    "uniserde",
                    "Pure Python, easy to use Serializer",
                    rio.URL("https://gitlab.com/Vivern/uniserde"),
                ),
            ),
            row_spacing=2,
            column_spacing=2,
        )



import watchfiles
import asyncio
from pathlib import Path




async def main() -> None:
    watch_path = Path.home() / "Local" / "rio"
    async for changes in watchfiles.awatch(watch_path):
        print(changes)


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from . import component_base

__all__ = [
    "TextInput",
    "TextInputChangeEvent",
    "TextInputConfirmEvent",
]


@dataclass
class TextInputChangeEvent:
    text: str


@dataclass
class TextInputConfirmEvent:
    text: str


class TextInput(component_base.KeyboardFocusableFundamentalComponent):
    """
    A user-editable text field.

    `TextInput` allows the user to enter a short text. The text can either be
    shown in plain text, or hidden when used for passwords or other sensitive
    information.

    Attributes:
        text: The text currently entered by the user.

        label: A short text to display next to the text input.

        prefix_text: A short text to display before the text input. Useful for
            displaying currency symbols or other prefixed units.

        suffix_text: A short text to display after the text input. Useful for
            displaying units, parts of e-mail addresses, and similar.

        is_secret: Whether the text should be hidden. Use this to hide sensitive
            information such as passwords.

        is_sensitive: Whether the text input should respond to user input.

        is_valid: Whether the current text is valid for your application. You
            can use this to signal to the user that their input needs to be
            changed.

        on_change: Triggered when the user changes the text.

        on_confirm: Triggered when the user explicitly confirms their input,
            such as by pressing the "Enter" key. You can use this to trigger
            followup actions, such as logging in or submitting a form.
    """

    text: str = ""
    _: KW_ONLY
    label: str = ""
    prefix_text: str = ""
    suffix_text: str = ""
    is_secret: bool = False
    is_sensitive: bool = True
    is_valid: bool = True
    on_change: rio.EventHandler[TextInputChangeEvent] = None
    on_confirm: rio.EventHandler[TextInputConfirmEvent] = None

    async def _on_state_update(self, delta_state: JsonDoc) -> None:
        # Trigger on_change event
        try:
            new_value = delta_state["text"]
        except KeyError:
            pass
        else:
            assert isinstance(new_value, str), new_value
            await self.call_event_handler(
                self.on_change,
                TextInputChangeEvent(new_value),
            )

        # Chain up
        await super()._on_state_update(delta_state)

    async def _on_message(self, msg: Any) -> None:
        # Listen for messages indicating the user has confirmed their input
        #
        # In addition to notifying the backend, these also include the input's
        # current value. This ensures any event handlers actually use the up-to
        # date value.
        assert isinstance(msg, dict), msg
        self.text = msg["text"]

        await self.call_event_handler(
            self.on_change,
            TextInputChangeEvent(self.text),
        )

        await self.call_event_handler(
            self.on_confirm,
            TextInputConfirmEvent(self.text),
        )

        # Refresh the session
        await self.session._refresh()


TextInput._unique_id = "TextInput-builtin"



from __future__ import annotations

import abc
import asyncio
import dataclasses
import inspect
import json
import typing
import weakref
from abc import abstractmethod
from collections import defaultdict
from dataclasses import KW_ONLY, dataclass
from typing import *  # type: ignore

import introspection
from typing_extensions import dataclass_transform
from uniserde import Jsonable, JsonDoc

import rio

from .. import common, event, global_state, inspection

__all__ = ["Component"]


JAVASCRIPT_SOURCE_TEMPLATE = """
%(js_source)s

if (%(js_class_name)s !== undefined) {
    window.componentClasses['%(cls_unique_id)s'] = %(js_class_name)s;
    window.childAttributeNames['%(cls_unique_id)s'] = %(child_attribute_names)s;
}
"""


CSS_SOURCE_TEMPLATE = """
const style = document.createElement('style');
style.innerHTML = %(escaped_css_source)s;
document.head.appendChild(style);
"""


T = TypeVar("T")
P = ParamSpec("P")


def make_default_factory_for_value(value: T) -> Callable[[], T]:
    def default_factory() -> T:
        return value

    default_factory.__name__ = default_factory.__qualname__ = f"return_{value!r}"

    return default_factory


async def call_component_handler_once(
    weak_component: weakref.ReferenceType[Component],
    handler: Callable,
) -> None:
    # Does the component still exist?
    component = weak_component()

    if component is None:
        return

    # Call the handler
    await component.call_event_handler(lambda: handler(component))
    await component.session._refresh()


async def _periodic_event_worker(
    weak_component: weakref.ReferenceType[Component],
    handler: Callable,
) -> None:
    while True:
        # Wait for the next tick
        await asyncio.sleep(handler._rio_event_periodic_interval_)  # type: ignore

        # Call the handler
        await call_component_handler_once(weak_component, handler)


@dataclass(eq=False)
class StateBinding:
    # Weak reference to the component containing this binding
    owning_component_weak: Callable[[], Optional[Component]]

    # The state property whose value this binding is
    owning_property: StateProperty

    # Each binding is either the root-most binding, or a child of another
    # binding. This value is True if this binding is the root.
    is_root: bool

    parent: Optional[StateBinding]
    value: Optional[object]

    children: weakref.WeakSet[StateBinding] = dataclasses.field(
        default_factory=weakref.WeakSet
    )

    def get_value(self) -> object:
        if self.is_root:
            return self.value

        assert self.parent is not None
        return self.parent.get_value()

    def set_value(self, value: object) -> None:
        # Delegate to the parent, if any
        if self.parent is not None:
            self.parent.set_value(value)
            return

        # Otherwise this is the root-most binding. Set the value
        self.value = value

        # Then recursively mark all children as dirty
        self.recursively_mark_children_as_dirty()

    def recursively_mark_children_as_dirty(self) -> None:
        to_do = [self]

        while to_do:
            cur = to_do.pop()
            owning_component = cur.owning_component_weak()

            # The component's session may be `None`, if this component has never
            # entered the component tree. e.g. `build` returns a component, which
            # doesn't make it through reconciliation and is thus never even
            # marked as dirty -> No session is injected.
            if owning_component is not None and owning_component._session_ is not None:
                owning_component._session_._register_dirty_component(
                    owning_component,
                    include_children_recursively=False,
                )

            to_do.extend(cur.children)


class StateProperty:
    """
    StateProperties act like regular properties, with additional considerations:

    - When a state property is assigned to, the component owning it is marked as
      dirty in the session

    - State properties have the ability to share their value with other state
      property instances. If state property `A` is assigned to state property
      `B`, then `B` creates a `StateBinding` and any future access to `B` will
      be routed to `A` instead:

    ```
    class Foo(Component):
        foo_text = "Hello"

        def build(self) -> Component:
            return Bar(bar_text=Foo.foo_text)  # Note `Foo` instead of `self`
    ```
    """

    def __init__(self, name: str, readonly: bool):
        self.name = name
        self.readonly = readonly

    def __get__(
        self,
        instance: Optional[Component],
        owner: Optional[type] = None,
    ) -> object:
        # If accessed through the class, rather than instance, return the
        # StateProperty itself
        if instance is None:
            return self

        # Otherwise get the value assigned to the property in the component
        # instance
        try:
            value = vars(instance)[self.name]
        except KeyError:
            raise AttributeError(self.name) from None

        # If the value is a binding return the binding's value
        if isinstance(value, StateBinding):
            return value.get_value()

        # Otherwise return the value
        return value

    def __set__(self, instance: Component, value: object) -> None:
        if self.readonly:
            cls_name = type(instance).__name__
            raise AttributeError(
                f"Cannot assign to readonly property {cls_name}.{self.name}"
            )

        # Look up the stored value
        instance_vars = vars(instance)
        try:
            local_value = instance_vars[self.name]
        except KeyError:
            pass
        else:
            # If a value is already stored, that means this is a re-assignment.
            # Which further means it's an assignment outside of `__init__`.
            # Which is not a valid place to create a state binding.
            if isinstance(value, StateProperty):
                raise RuntimeError(
                    "State bindings can only be created when calling the component constructor"
                )

            # Delegate to the binding if it exists
            if isinstance(local_value, StateBinding):
                local_value.set_value(value)
                return

        # Otherwise set the value directly and mark the component as dirty
        instance_vars[self.name] = value

        if instance._session_ is not None:
            instance._session_._register_dirty_component(
                instance,
                include_children_recursively=False,
            )

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self.name}>"


class ComponentMeta(abc.ABCMeta):
    def __call__(cls, *args: object, **kwargs: object):
        component = super().__call__(*args, **kwargs)
        component._create_state_bindings()

        # Call `_rio_post_init` for every class in the MRO
        for base in reversed(type(component).__mro__):
            try:
                post_init = vars(base)["_rio_post_init"]
            except KeyError:
                continue

            post_init(component)

        # Call `on_create` handlers
        on_create_handlers = component._rio_event_handlers_[event.EventTag.ON_CREATE]

        for handler in on_create_handlers:
            handler(component)

        return component


@dataclass_transform(eq_default=False)
@dataclass(eq=False, repr=False)
class Component(metaclass=ComponentMeta):
    """
    Base class for all `rio` components.

    Components are the building blocks of `rio` apps. `rio` ships with many
    useful components out of the box, but you can also subclass a component to
    create your own.

    Attributes:
        key: A unique identifier for this component. If two components with the
            same key are present during reconciliation they will be considered
            the same component and their state will be preserved. If no key is
            specified, reconciliation falls back to a less precise method, by
            comparing the location of the component in the component tree.

        margin: The margin around this component. This is a shorthand for
            setting `margin_left`, `margin_top`, `margin_right` and
            `margin_bottom` to the same value. If multiple conflicting margins
            are specified the most specific one wins. If for example `margin`
            and `margin_left` are both specified, `margin_left` is used for the
            left side, while the other sides use `margin`.

        margin_x: The horizontal margin around this component. This is a
            shorthand for setting `margin_left` and `margin_right` to the same
            value. If multiple conflicting margins are specified the most
            specific one wins. If for example `margin_x` and `margin_left` are
            both specified, `margin_left` is used for the left side, while the
            other side uses `margin_x`.

        margin_y: The vertical margin around this component. This is a shorthand
            for setting `margin_top` and `margin_bottom` to the same value. If
            multiple conflicting margins are specified the most specific one
            wins. If for example `margin_y` and `margin_top` are both specified,
            `margin_top` is used for the top side, while the other side uses
            `margin_y`.

        margin_left: The left margin around this component. If multiple
            conflicting margins are specified this one will be used, since it's
            the most specific. If for example `margin_left` and `margin` are
            both specified, `margin_left` is used for the left side, while the
            other sides use `margin`.

        margin_top: The top margin around this component. If multiple
            conflicting margins are specified this one will be used, since it's
            the most specific. If for example `margin_top` and `margin` are both
            specified, `margin_top` is used for the top side, while the other
            sides use `margin`.

        margin_right: The right margin around this component. If multiple
            conflicting margins are specified this one will be used, since it's
            the most specific. If for example `margin_right` and `margin` are
            both specified, `margin_right` is used for the right side, while the
            other sides use `margin`.

        margin_bottom: The bottom margin around this component. If multiple
            conflicting margins are specified this one will be used, since it's
            the most specific. If for example `margin_bottom` and `margin` are
            both specified, `margin_bottom` is used for the bottom side, while
            the other sides use `margin`.

        width: How much horizontal space this component should request during
            layouting. This can be either a number, or one of the special
            values:

            If `"natural"`, the component will request the minimum amount it
            requires to fit on the screen. For example a `Text` will request
            however much space the characters of that text require. A `Row`
            would request the sum of the widths of its children.

            If `"grow"`, the component will request all the remaining space in
            its parent.

            Please note that the space a `Component` receives during layouting
            may not match the request. As a general rule for example, containers
            try to pass on all available space to children. If you really want a
            `Component` to only take up as much space as requested, consider
            specifying an alignment.

        height: How much vertical space this component should request during
            layouting. This can be either a number, or one of the special
            values:

            If `"natural"`, the component will request the minimum amount it
            requires to fit on the screen. For example a `Text` will request
            however much space the characters of that text require. A `Row`
            would request the height of its tallest child.

            If `"grow"`, the component will request all the remaining space in
            its parent.

            Please note that the space a `Component` receives during layouting
            may not match the request. As a general rule for example, containers
            try to pass on all available space to children. If you really want a
            `Component` to only take up as much space as requested, consider
            specifying an alignment.

        align_x: How this component should be aligned horizontally, if it
            receives more space than it requested. This can be a number between
            0 and 1, where 0 means left-aligned, 0.5 means centered, and 1 means
            right-aligned.

        align_y: How this component should be aligned vertically, if it receives
            more space than it requested. This can be a number between 0 and 1,
            where 0 means top-aligned, 0.5 means centered, and 1 means
            bottom-aligned.
    """

    _: KW_ONLY
    key: Optional[str] = None

    margin: Optional[float] = None
    margin_x: Optional[float] = None
    margin_y: Optional[float] = None

    margin_left: Optional[float] = None
    margin_top: Optional[float] = None
    margin_right: Optional[float] = None
    margin_bottom: Optional[float] = None

    width: Union[Literal["natural", "grow"], float] = "natural"
    height: Union[Literal["natural", "grow"], float] = "natural"

    align_x: Optional[float] = None
    align_y: Optional[float] = None

    _id: int = dataclasses.field(init=False)

    # Weak reference to the component's builder. Used to check if the component is
    # still part of the component tree.
    _weak_builder_: Callable[[], Optional[Component]] = dataclasses.field(
        # Dataclasses seem to unintentionally turn this function into a method.
        # Make sure it works whether or not `self` is passed.
        default=lambda *args: None,
        init=False,
    )

    # Each time a component is built the build generation in that component's COMPONENT
    # DATA is incremented. If this value no longer matches the value in its
    # builder's COMPONENT DATA, the component is dead.
    _build_generation_: int = dataclasses.field(default=-1, init=False)

    # Injected by the session when the component is refreshed
    _session_: Optional["rio.Session"] = dataclasses.field(default=None, init=False)

    # Remember which properties were explicitly set in the constructor. This is
    # filled in by `__new__`
    _explicitly_set_properties_: Set[str] = dataclasses.field(
        init=False, default_factory=set
    )

    # True if the component is subscribed to either the `on_mount` or
    # `on_unmount` events.
    _watch_tree_mount_and_unmount_: ClassVar[bool] = dataclasses.field(
        default=False, init=False
    )

    # This flag indicates whether state bindings for this component have already
    # been initialized. Used by `__getattribute__` to check if it should throw
    # an error.
    _state_bindings_initialized_: bool = dataclasses.field(default=False, init=False)

    # Cache for the set of all `StateProperty` instances in this class
    _state_properties_: ClassVar[Dict[str, "StateProperty"]]

    # Maps event tags to the methods that handle them. The methods aren't bound
    # to the instance yet, so make sure to pass `self` when calling them
    #
    # The assigned value is needed so that the `Component` class itself has a
    # valid value. All subclasses override this value in `__init_subclass__`.
    _rio_event_handlers_: ClassVar[
        defaultdict[event.EventTag, Tuple[Callable[..., Any], ...]]
    ] = defaultdict(tuple)

    @classmethod
    def _preprocess_dataclass_fields(cls):
        # When a field has a default value (*not* default factory!), the
        # constructor actually doesn't assign the default value to the instance.
        # The default value is actually permanently stored in the class. So the
        # instance doesn't have the attribute, but the class does, and
        # everything is fine, right? Wrong. We create a `StateProperty` for each
        # field, which overrides that default value. We absolutely need every
        # attribute to be an instance attribute, which we can achieve by
        # replacing every default value with a default factory.
        cls_vars = vars(cls)

        for attr_name in cls_vars.get("__annotations__", {}):
            try:
                field_or_default = cls_vars[attr_name]
            except KeyError:
                continue

            if not isinstance(field_or_default, dataclasses.Field):
                field = dataclasses.field(
                    default_factory=make_default_factory_for_value(field_or_default)
                )
                setattr(cls, attr_name, field)
                continue

            field: dataclasses.Field[object] = field_or_default

            # If it doesn't have a default value, we can ignore it
            if field.default is dataclasses.MISSING:
                continue

            field.default_factory = make_default_factory_for_value(field.default)
            field.default = dataclasses.MISSING

    @staticmethod
    def _determine_explicitly_set_properties(
        original_init: Callable[Concatenate[Component, P], None],
        self: "Component",
        *args: P.args,
        **kwargs: P.kwargs,
    ):
        # Chain up to the original `__init__`
        original_init(self, *args, **kwargs)

        # Determine which properties were explicitly set
        bound_args = inspect.signature(original_init).bind(self, *args, **kwargs)
        self._explicitly_set_properties_.update(bound_args.arguments)

    @staticmethod
    def _init_component(
        original_init: Callable[Concatenate[Component, P], None],
        self: "Component",
        *args: P.args,
        **kwargs: P.kwargs,
    ):
        # Fetch the session this component is part of
        if global_state.currently_building_session is None:
            raise RuntimeError(
                "Components can only be created inside of `build` methods."
            )

        session = global_state.currently_building_session
        self._session_ = session

        # Create a unique ID for this component
        self._id = session._next_free_component_id
        session._next_free_component_id += 1

        session._register_dirty_component(
            self,
            include_children_recursively=False,
        )

        # Keep track of this component's existence
        #
        # Components must be known by their id, so any messages addressed to
        # them can be passed on correctly.
        session._weak_components_by_id[self._id] = self

        # Some events need attention right after the component is created
        for event_tag, event_handlers in self._rio_event_handlers_.items():
            # Don't register event handlers if there aren't any, since that would slow down the session
            if not event_handlers:
                continue

            # Page changes are handled by the session. Register the handler
            if event_tag == event.EventTag.ON_PAGE_CHANGE:
                session._page_change_callbacks[self] = event_handlers

            # The `periodic` event needs a task to work in
            elif event_tag == event.EventTag.PERIODIC:
                for event_handler in event_handlers:
                    session.create_task(
                        _periodic_event_worker(weakref.ref(self), event_handler),
                        name=f"`rio.event.periodic` event worker for {self}",
                    )

        # Call the `__init__` created by `@dataclass`
        original_init(self, *args, **kwargs)

        # Initialize the margins. This has to happen after the dataclass
        # `__init__`, because it would overwrite our values.
        def elvis(*param_names: str) -> Any:
            for param_name in param_names:
                value = kwargs.get(param_name)

                if value is not None:
                    return value

            return 0

        self.margin_left = elvis("margin_left", "margin_x", "margin")
        self.margin_top = elvis("margin_top", "margin_y", "margin")
        self.margin_right = elvis("margin_right", "margin_x", "margin")
        self.margin_bottom = elvis("margin_bottom", "margin_y", "margin")

    def _create_state_bindings(self) -> None:
        self._state_bindings_initialized_ = True

        creator = global_state.currently_building_component

        # The creator can be `None` if this component was created by the app's
        # `build` function. It's not possible to create a state binding in that
        # case.
        if creator is None:
            return

        creator_vars = vars(creator)
        self_vars = vars(self)

        for prop_name, state_property in self._state_properties_.items():
            # The dataclass constructor doesn't actually assign default values
            # as instance attributes. In that case we get a KeyError here. But
            # default values can't be state bindings, so just skip this
            # attribute
            try:
                value = self_vars[prop_name]
            except KeyError:
                continue

            # Create a StateBinding, if requested
            if not isinstance(value, StateProperty):
                continue

            # In order to create a `StateBinding`, the creator's
            # attribute must also be a binding
            parent_binding = creator_vars[value.name]

            if not isinstance(parent_binding, StateBinding):
                parent_binding = StateBinding(
                    owning_component_weak=weakref.ref(creator),
                    owning_property=value,
                    is_root=True,
                    parent=None,
                    value=parent_binding,
                    children=weakref.WeakSet(),
                )
                creator_vars[value.name] = parent_binding

            # Create the child binding
            child_binding = StateBinding(
                owning_component_weak=weakref.ref(self),
                owning_property=state_property,
                is_root=False,
                parent=parent_binding,
                value=None,
                children=weakref.WeakSet(),
            )
            parent_binding.children.add(child_binding)
            self_vars[prop_name] = child_binding

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        cls_vars = vars(cls)
        has_custom_init = "__init__" in cls_vars

        # We'll manually take care of calling __post_init__ methods because
        # dataclasses are unreliable. Rename the method so that it won't be
        # called by the dataclass constructor.
        if "__post_init__" in cls_vars:
            cls._rio_post_init = cls.__post_init__  # type: ignore
            del cls.__post_init__  # type: ignore

        # Apply the dataclass transform
        cls._preprocess_dataclass_fields()
        dataclasses.dataclass(eq=False, repr=False)(cls)

        # Keep track of which properties were explicitly set in the constructor.
        introspection.wrap_method(
            cls._determine_explicitly_set_properties,
            cls,
            "__init__",
        )

        # Components need to run custom code in in `__init__`, but dataclass
        # constructors don't chain up. So if this class's `__init__` was created
        # by the `@dataclass` decorator, wrap it with a custom `__init__` that
        # calls our initialization code.
        if not has_custom_init:
            introspection.wrap_method(
                cls._init_component,
                cls,
                "__init__",
            )

        # Replace all properties with custom state properties
        all_parent_state_properties: Dict[str, StateProperty] = {}

        for base in cls.__bases__:
            if not issubclass(base, Component):
                continue

            all_parent_state_properties.update(base._state_properties_)

        cls._initialize_state_properties(all_parent_state_properties)

        # Keep track of all event handlers. By gathering them here, the
        # component constructor doesn't have to re-scan the entire class for
        # each instantiation.
        event_handlers: defaultdict[
            event.EventTag, List[Callable[..., Any]]
        ] = defaultdict(list)

        # Inherit event handlers from base classes
        for base in cls.__bases__:
            if not issubclass(base, Component):
                continue

            for event_tag, handlers in base._rio_event_handlers_.items():
                event_handlers[event_tag].extend(handlers)

        # Add any event handlers added in this class
        for method in vars(cls).values():
            try:
                tag = method._rio_event_tag_
            except AttributeError:
                continue

            event_handlers[tag].append(method)

        # Convert the lists to tuples
        cls._rio_event_handlers_ = defaultdict(
            tuple,
            {
                event_tag: tuple(handlers)
                for event_tag, handlers in event_handlers.items()
            },
        )

        # If the component is subscribed to `on_mount` or `on_unmount`, update
        # the boolean for that
        if (
            event.EventTag.ON_MOUNT in cls._rio_event_handlers_
            or event.EventTag.ON_UNMOUNT in cls._rio_event_handlers_
        ):
            cls._watch_tree_mount_and_unmount_ = True

    @classmethod
    def _initialize_state_properties(
        cls,
        parent_state_properties: Dict[str, "StateProperty"],
    ) -> None:
        """
        Spawn `StateProperty` instances for all annotated properties in this
        class.
        """
        cls._state_properties_ = parent_state_properties.copy()

        # TODO: Placeholder function, until a better one is implemented in the
        # `introspection` package.
        def is_classvar(annotation: Any) -> bool:
            if isinstance(annotation, str):
                return annotation.startswith(("ClassVar", "typing.ClassVar"))

            return typing.get_origin(annotation) is ClassVar

        for attr_name, annotation in vars(cls).get("__annotations__", {}).items():
            # Skip `ClassVar` annotations
            if is_classvar(annotation):
                continue

            # Skip internal properties. These aren't supposed to be wrapped in
            # `StateProperty`
            if attr_name in (
                "_",
                "_build_generation_",
                "_explicitly_set_properties_",
                "_id",
                "_init_signature_",
                "_session_",
                "_weak_builder_",
                "_state_bindings_initialized_",
                "margin_x",
                "margin_y",
                "margin",
            ):
                continue

            # Create the `StateProperty`
            # readonly = introspection.typing.has_annotation(annotation, Readonly
            readonly = False  # FIXME
            state_property = StateProperty(attr_name, readonly)
            setattr(cls, attr_name, state_property)

            # Add it to the set of all state properties for rapid lookup
            cls._state_properties_[attr_name] = state_property

    # When running in development mode, make sure that no component `__init__`
    # tries to read a state property. This would be incorrect because state
    # bindings are not yet initialized at that point.
    if common.RUNNING_IN_DEV_MODE and not TYPE_CHECKING:

        def __getattribute__(self, attr_name: str):
            # fmt: off
            if (
                attr_name in type(self)._state_properties_ and
                not super().__getattribute__("_state_bindings_initialized_")
            ):
                # fmt: on
                raise Exception(
                    "You have attempted to read a state property in a component's"
                    " `__init__` method. This is not allowed because state"
                    " bindings are not yet initialized at that point. Please"
                    " move this code into the `on_create` method."
                )

            return super().__getattribute__(attr_name)

    @property
    def session(self) -> "rio.Session":
        """
        Return the session this component is part of.

        The session is accessible after the build method which constructed this
        component has returned.
        """
        if self._session_ is None:
            raise RuntimeError(
                "The session is only accessible once the build method which"
                " constructed this component has returned."
            )

        return self._session_

    def _custom_serialize(self) -> JsonDoc:
        """
        Return any additional properties to be serialized, which cannot be
        deduced automatically from the type annotations.
        """
        return {}

    @abstractmethod
    def build(self) -> "Component":
        """
        Return a component tree which represents the UI of this component.

        Most components define their appearance and behavior by combining other,
        more basic components. This function's purpose is to do exactly that. It
        returns another component (typically a container) which will be
        displayed on the screen.

        The `build` function should be pure, meaning that it does not modify the
        component's state and returns the same result each time it's invoked.
        """
        raise NotImplementedError()

    def _iter_direct_children(self) -> Iterable["Component"]:
        for name in inspection.get_child_component_containing_attribute_names(
            type(self)
        ):
            try:
                value = getattr(self, name)
            except AttributeError:
                continue

            if isinstance(value, Component):
                yield value

            if isinstance(value, list):
                value = cast(List[object], value)

                for item in value:
                    if isinstance(item, Component):
                        yield item

    def _iter_direct_and_indirect_child_containing_attributes(
        self,
        *,
        include_self: bool,
        recurse_into_high_level_components: bool,
    ) -> Iterable["Component"]:
        # Special case the component itself to handle `include_self`
        if include_self:
            yield self

        if not recurse_into_high_level_components and not isinstance(
            self, FundamentalComponent
        ):
            return

        # Iteratively yield all children
        to_do = list(self._iter_direct_children())
        while to_do:
            cur = to_do.pop()
            yield cur

            if recurse_into_high_level_components or isinstance(
                cur, FundamentalComponent
            ):
                to_do.extend(cur._iter_direct_children())

    def _iter_component_tree(self) -> Iterable["Component"]:
        """
        Iterate over all components in the component tree, with this component as the root.
        """
        yield self

        if isinstance(self, FundamentalComponent):
            for child in self._iter_direct_children():
                yield from child._iter_component_tree()
        else:
            build_result = self.session._weak_component_data_by_component[
                self
            ].build_result
            yield from build_result._iter_component_tree()

    async def _on_message(self, msg: Jsonable, /) -> None:
        raise RuntimeError(f"{type(self).__name__} received unexpected message `{msg}`")

    def _is_in_component_tree(self, cache: Dict[rio.Component, bool]) -> bool:
        """
        Returns whether this component is directly or indirectly connected to the
        component tree of a session.

        This operation is fast, but has to walk up the component tree to make sure
        the component's parent is also connected. Thus, when checking multiple
        components it can easily happen that the same components are checked over and
        over, resulting on O(n log n) runtime. To avoid this, pass a cache
        dictionary to this function, which will be used to memoize the result.

        Be careful not to reuse the cache if the component hierarchy might have
        changed (for example after an async yield).
        """

        # Already cached?
        try:
            return cache[self]
        except KeyError:
            pass

        # Root component?
        if self is self.session._root_component:
            result = True

        # Has the builder has been garbage collected?
        else:
            builder = self._weak_builder_()
            if builder is None:
                result = False

            # Has the builder since created new build output, and this component
            # isn't part of it anymore?
            else:
                parent_data = self.session._weak_component_data_by_component[builder]
                result = (
                    parent_data.build_generation == self._build_generation_
                    and builder._is_in_component_tree(cache)
                )

        # Cache the result and return
        cache[self] = result
        return result

    @overload
    async def call_event_handler(
        self,
        handler: rio.EventHandler[[]],
    ) -> None:
        ...

    @overload
    async def call_event_handler(
        self,
        handler: rio.EventHandler[[T]],
        event_data: T,
        /,
    ) -> None:
        ...

    async def call_event_handler(
        self,
        handler: rio.EventHandler[...],
        *event_data: object,
    ) -> None:
        """
        Calls an even handler, awaiting it if necessary.

        Call an event handler, if one is present. Await it if necessary. Log and
        discard any exceptions. If `event_data` is present, it will be passed to
        the event handler.
        """
        await self.session._call_event_handler(handler, *event_data)

    async def force_refresh(self) -> None:
        """
        Force a rebuild of this component.

        Most of the time components update automatically when their state
        changes. However, some state mutations are invisible to `Rio`: For
        example, appending items to a list modifies the list, but since no list
        instance was actually assigned to th component, `Rio` will be unaware of
        this change.

        In these cases, you can force a rebuild of the component by calling
        `force_refresh`. This will trigger a rebuild of the component and
        display the updated version on the screen.

        Another common use case is if you wish to update an component while an
        event handler is still running. `Rio` will automatically detect changes
        after event handlers return, but if you are performing a long-running
        operation, you may wish to update the component while the event handler
        is still running. This allows you to e.g. update a progress bar while
        the operation is still running.
        """
        self.session._register_dirty_component(
            self,
            include_children_recursively=False,
        )

        await self.session._refresh()

    def __repr__(self) -> str:
        result = f"<{type(self).__name__} id:{self._id}"

        child_strings: List[str] = []
        for child in self._iter_direct_children():
            child_strings.append(f" {type(child).__name__}:{child._id}")

        if child_strings:
            result += " -" + "".join(child_strings)

        return result + ">"


# Most classes have their state properties initialized in
# `Component.__init_subclass__`. However, since `Component` isn't a subclass of
# itself this needs to be done manually.
Component._initialize_state_properties({})
introspection.wrap_method(
    Component._init_component,
    Component,
    "__init__",
)


class FundamentalComponent(Component):
    # Unique id for identifying this class in the frontend. This is initialized
    # in `Component.__init_subclass__`.
    _unique_id: ClassVar[str]

    def build(self) -> "Component":
        raise RuntimeError(
            f"Attempted to call `build` on `FundamentalComponent` {self}"
        )

    @classmethod
    def build_javascript_source(cls, sess: rio.Session) -> str:
        return ""

    @classmethod
    def build_css_source(cls, sess: rio.Session) -> str:
        return ""

    def __init_subclass__(cls):
        # Assign a unique id to this class. This allows the frontend to identify
        # components.
        hash_ = common.secure_string_hash(
            cls.__module__,
            cls.__qualname__,
            hash_length=12,
        )

        cls._unique_id = f"{cls.__name__}-{hash_}"

        # Chain up
        super().__init_subclass__()

    @classmethod
    async def _initialize_on_client(cls, sess: rio.Session) -> None:
        message_source = ""

        javascript_source = cls.build_javascript_source(sess)
        if javascript_source:
            message_source += JAVASCRIPT_SOURCE_TEMPLATE % {
                "js_source": javascript_source,
                "js_class_name": cls.__name__,
                "cls_unique_id": cls._unique_id,
                "child_attribute_names": json.dumps(
                    inspection.get_child_component_containing_attribute_names(cls)
                ),
            }

        css_source = cls.build_css_source(sess)
        if css_source:
            escaped_css_source = json.dumps(css_source)
            message_source += CSS_SOURCE_TEMPLATE % {
                "escaped_css_source": escaped_css_source,
            }

        if message_source:
            await sess._evaluate_javascript(message_source)

    async def _on_state_update(self, delta_state: JsonDoc) -> None:
        """
        This function is called when the frontend sends a state update to this
        component.
        """
        # Update all state properties to reflect the new state
        for attr_name, attr_value in delta_state.items():
            assert isinstance(attr_value, (bool, int, float, str)), attr_value
            assert hasattr(type(self), attr_name), attr_name
            assert isinstance(getattr(type(self), attr_name), StateProperty), attr_name

            setattr(self, attr_name, attr_value)

        # Trigger a refresh
        assert self._session_ is not None
        await self._session_._refresh()

    async def _on_message(self, message: Jsonable, /) -> None:
        """
        This function is called when the frontend sends a message to this component
        via `sendMessage`.
        """
        pass


class KeyboardFocusableFundamentalComponent(FundamentalComponent):
    async def grab_keyboard_focus(self) -> None:
        await self.session._remote_set_keyboard_focus(self._id)
